# Copyright 2015-2016, Tresys Technology, LLC
# Copyright 2018, Chris PeBenito <pebenito@ieee.org>
#
# SPDX-License-Identifier: LGPL-2.1-only
#
import logging
from abc import ABC, abstractmethod
from typing import Generic, Iterable, TypeVar

from ..policyrep import PolicyObject, PolicySymbol, SELinuxPolicy


class Difference:

    """Base class for all policy differences."""

    def __init__(self, left_policy: SELinuxPolicy, right_policy: SELinuxPolicy) -> None:
        self.log = logging.getLogger(__name__)
        self.left_policy = left_policy
        self.right_policy = right_policy

    #
    # Policies to compare
    #
    @property
    def left_policy(self):
        return self._left_policy

    @left_policy.setter
    def left_policy(self, policy):
        self.log.info("Policy diff left policy set to {0}".format(policy))
        self._left_policy = policy
        self._reset_diff()

    @property
    def right_policy(self):
        return self._right_policy

    @right_policy.setter
    def right_policy(self, policy):
        self.log.info("Policy diff right policy set to {0}".format(policy))
        self._right_policy = policy
        self._reset_diff()

    #
    # Internal functions
    #
    def _reset_diff(self) -> None:
        """Reset diff results on policy changes."""
        raise NotImplementedError

    @staticmethod
    def _expand_generator(rule_list: Iterable, wrapper_class) -> Iterable:
        """Generator that yields a wrapped, expanded rule list."""
        # this is to delay creating any containers
        # as long as possible, since rule lists
        # are typically massive.
        for unexpanded_rule in rule_list:
            for expanded_rule in unexpanded_rule.expand():
                yield wrapper_class(expanded_rule)

    @staticmethod
    def _set_diff(left, right, key=None, unwrap=True):
        """
        Standard diff of two sets.

        Parameters:
        left        An iterable
        right       An iterable

        Return:
        tuple       (added, removed, matched)

        added       Set of items in right but not left
        removed     Set of items in left but not right
        matched     Set of items in both left and right.  This is
                    in the form of tuples with the matching item
                    from left and right
        """

        left_items = set(left)
        right_items = set(right)
        added_items = right_items - left_items
        removed_items = left_items - right_items

        # The problem here is the symbol from both policies are
        # needed to build each tuple in the matched items set.
        # Using the standard Python set intersection code will only result
        # in one object.
        #
        # This tuple-generating code creates lists from the sets, to sort them.
        # This should result in all of the symbols lining up.  If they don't,
        # this will break the caller.  This should work since there is no remapping.
        #
        # This has extra checking to make sure this assertion holds, to fail
        # instead of giving wrong results.  If there is a better way to,
        # ensure the items match up, please let me know how or submit a patch.
        matched_items = set()
        left_matched_items = sorted((left_items - removed_items), key=key)
        right_matched_items = sorted((right_items - added_items), key=key)
        assert len(left_matched_items) == len(right_matched_items), \
            "Matched items assertion failure (this is an SETools bug), {0} != {1}". \
            format(len(left_matched_items), len(right_matched_items))

        for left, right in zip(left_matched_items, right_matched_items):
            assert left == right, \
                "Matched items assertion failure (this is an SETools bug), {0} != {1}".format(
                    left, right)

            matched_items.add((left, right))

        if unwrap:
            return set(i.origin for i in added_items), \
                set(i.origin for i in removed_items), \
                set((left.origin, right.origin) for (left, right) in matched_items)
        else:
            return added_items, removed_items, matched_items


T = TypeVar("T", bound=PolicyObject)


class Wrapper(ABC, Generic[T]):

    """Abstract base class for policy object wrappers."""

    origin: T
    key: int

    __slots__ = ("origin", "key")

    def __init__(self, origin: T, key: int) -> None:
        self.origin = origin
        self.key = key

    def __repr__(self):
        # pylint: disable=no-member
        return "<{0.__class__.__name__}(Wrapping {1})>".format(self, repr(self.origin))

    @abstractmethod
    def __hash__(self):
        pass

    @abstractmethod
    def __eq__(self, other):
        pass

    @abstractmethod
    def __lt__(self, other):
        pass

    def __ne__(self, other):
        return not self == other


S = TypeVar("S", bound=PolicySymbol)


class SymbolWrapper(Wrapper[S]):

    """
    General wrapper for policy symbols, e.g. types, roles
    to provide a diff-specific equality operation based
    on its name.
    """

    name: str

    __slots__ = ("name",)

    def __init__(self, symbol: S) -> None:
        super().__init__(symbol, hash(symbol.name))
        self.name = symbol.name

    def __hash__(self):
        return self.key

    def __lt__(self, other):
        return self.name < other.name

    def __eq__(self, other):
        return self.name == other.name



# Copyright 2016, Tresys Technology, LLC
#
# SPDX-License-Identifier: LGPL-2.1-only
#
from typing import NamedTuple, Union

from ..policyrep import PolicyEnum

from .descriptors import DiffResultDescriptor
from .difference import Difference


class ModifiedProperty(NamedTuple):

    """Difference details for a modified policy property."""

    property: str
    added: Union[PolicyEnum, bool, int]
    removed: Union[PolicyEnum, bool, int]


class PropertiesDifference(Difference):

    """
    Determine the difference in policy properties
    (unknown permissions, MLS, etc.) between two policies.
    """

    modified_properties = DiffResultDescriptor("diff_properties")

    def diff_properties(self) -> None:
        self.modified_properties = []

        if self.left_policy.handle_unknown != self.right_policy.handle_unknown:
            self.modified_properties.append(
                ModifiedProperty("handle_unknown",
                                 self.right_policy.handle_unknown,
                                 self.left_policy.handle_unknown))

        if self.left_policy.mls != self.right_policy.mls:
            self.modified_properties.append(
                ModifiedProperty("MLS",
                                 self.right_policy.mls,
                                 self.left_policy.mls))

        if self.left_policy.version != self.right_policy.version:
            self.modified_properties.append(
                ModifiedProperty("version",
                                 self.right_policy.version,
                                 self.left_policy.version))

    #
    # Internal functions
    #
    def _reset_diff(self) -> None:
        """Reset diff results on policy changes."""
        self.log.debug("Resetting property differences")
        self.modified_properties = None



# Copyright 2020, Microsoft Corporation
#
# SPDX-License-Identifier: LGPL-2.1-only
#

# This is a separate file to break a circular import.
CHECK_TYPE_KEY = "check_type"
CHECK_DESC_KEY = "desc"
CHECK_DISABLE = "disable"

GLOBAL_CONFIG_KEYS = frozenset((CHECK_TYPE_KEY, CHECK_DESC_KEY, CHECK_DISABLE))



# Copyright 2018, Chris PeBenito <pebenito@ieee.org>
#
# SPDX-License-Identifier: LGPL-2.1-only
#
import logging
from typing import Iterable, Optional

from .mixins import MatchContext, MatchName
from .policyrep import Ibendportcon
from .query import PolicyQuery
from .util import match_regex


class IbendportconQuery(MatchContext, MatchName, PolicyQuery):

    """
    Infiniband endport context query.

    Parameter:
    policy          The policy to query.

    Keyword Parameters/Class attributes:
    name            The name of the network interface to match.
    name_regex      If true, regular expression matching will
                    be used for matching the name.
    port            The port number to match.
    user            The criteria to match the context's user.
    user_regex      If true, regular expression matching
                    will be used on the user.
    role            The criteria to match the context's role.
    role_regex      If true, regular expression matching
                    will be used on the role.
    type_           The criteria to match the context's type.
    type_regex      If true, regular expression matching
                    will be used on the type.
    range_          The criteria to match the context's range.
    range_subset    If true, the criteria will match if it is a subset
                    of the context's range.
    range_overlap   If true, the criteria will match if it overlaps
                    any of the context's range.
    range_superset  If true, the criteria will match if it is a superset
                    of the context's range.
    range_proper    If true, use proper superset/subset operations.
                    No effect if not using set operations.
    """

    _port: Optional[int] = None

    @property
    def port(self) -> Optional[int]:
        return self._port

    @port.setter
    def port(self, value: Optional[int]) -> None:
        if value:
            pending_value = int(value)
            if not 0 < pending_value < 256:
                raise ValueError("Endport value must be 1-255.")

            self._port = pending_value
        else:
            self._port = None

    def __init__(self, policy, **kwargs):
        super(IbendportconQuery, self).__init__(policy, **kwargs)
        self.log = logging.getLogger(__name__)

    def results(self) -> Iterable[Ibendportcon]:
        """Generator which yields all matching ibendportcons."""
        self.log.info("Generating ibendportcon results from {0.policy}".format(self))
        self._match_name_debug(self.log)
        self.log.debug("Port: {0.port!r}".format(self))
        self._match_context_debug(self.log)

        for endport in self.policy.ibendportcons():
            if self.name and not match_regex(
                    endport.name,
                    self.name,
                    self.name_regex):
                continue

            if self.port is not None and self.port != endport.port:
                continue

            if not self._match_context(endport.context):
                continue

            yield endport



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.


import typing

from cryptography import utils
from cryptography.exceptions import (
    AlreadyFinalized,
)
from cryptography.hazmat.primitives import ciphers


class CMAC(object):
    def __init__(
        self,
        algorithm: ciphers.BlockCipherAlgorithm,
        backend: typing.Any = None,
        ctx=None,
    ):
        if not isinstance(algorithm, ciphers.BlockCipherAlgorithm):
            raise TypeError("Expected instance of BlockCipherAlgorithm.")
        self._algorithm = algorithm

        if ctx is None:
            from cryptography.hazmat.backends.openssl.backend import (
                backend as ossl,
            )

            self._ctx = ossl.create_cmac_ctx(self._algorithm)
        else:
            self._ctx = ctx

    def update(self, data: bytes) -> None:
        if self._ctx is None:
            raise AlreadyFinalized("Context was already finalized.")

        utils._check_bytes("data", data)
        self._ctx.update(data)

    def finalize(self) -> bytes:
        if self._ctx is None:
            raise AlreadyFinalized("Context was already finalized.")
        digest = self._ctx.finalize()
        self._ctx = None
        return digest

    def verify(self, signature: bytes) -> None:
        utils._check_bytes("signature", signature)
        if self._ctx is None:
            raise AlreadyFinalized("Context was already finalized.")

        ctx, self._ctx = self._ctx, None
        ctx.verify(signature)

    def copy(self) -> "CMAC":
        if self._ctx is None:
            raise AlreadyFinalized("Context was already finalized.")
        return CMAC(self._algorithm, ctx=self._ctx.copy())



# Copyright 2016, Tresys Technology, LLC
# Copyright 2018, Chris PeBenito <pebenito@ieee.org>
#
# SPDX-License-Identifier: LGPL-2.1-only
#
from typing import cast, List, NamedTuple, Optional

from ..policyrep import Bounds, BoundsRuletype, Type
from .descriptors import DiffResultDescriptor
from .difference import Difference, Wrapper
from .types import type_wrapper_factory


class ModifiedBounds(NamedTuple):

    """Difference details for a modified bounds rule."""

    rule: Bounds
    added_bound: Type
    removed_bound: Type


class BoundsDifference(Difference):

    """Determine the difference in *bounds between two policies."""

    added_typebounds = DiffResultDescriptor("diff_typebounds")
    removed_typebounds = DiffResultDescriptor("diff_typebounds")
    modified_typebounds = DiffResultDescriptor("diff_typebounds")

    # Lists of rules for each policy
    _left_typebounds: Optional[List[Bounds]] = None
    _right_typebounds: Optional[List[Bounds]] = None

    def diff_typebounds(self) -> None:
        """Generate the difference in typebound rules between the policies."""

        self.log.info("Generating typebounds differences from {0.left_policy} to {0.right_policy}".
                      format(self))

        if self._left_typebounds is None or self._right_typebounds is None:
            self._create_typebound_lists()

        self.added_typebounds, self.removed_typebounds, matched_typebounds = self._set_diff(
            (BoundsWrapper(c) for c in cast(List[Bounds], self._left_typebounds)),
            (BoundsWrapper(c) for c in cast(List[Bounds], self._right_typebounds)),
            key=lambda b: str(b.child))

        self.modified_typebounds = []

        for left_bound, right_bound in matched_typebounds:
            if type_wrapper_factory(left_bound.parent) != type_wrapper_factory(right_bound.parent):
                self.modified_typebounds.append(ModifiedBounds(
                    left_bound, right_bound.parent, left_bound.parent))

    #
    # Internal functions
    #
    def _create_typebound_lists(self) -> None:
        """Create rule lists for both policies."""
        self._left_typebounds = []
        for rule in self.left_policy.bounds():
            if rule.ruletype == BoundsRuletype.typebounds:
                self._left_typebounds.append(rule)
            else:
                self.log.error("Unknown rule type: {0} (This is an SETools bug)".
                               format(rule.ruletype))

        self._right_typebounds = []
        for rule in self.right_policy.bounds():
            if rule.ruletype == BoundsRuletype.typebounds:
                self._right_typebounds.append(rule)
            else:
                self.log.error("Unknown rule type: {0} (This is an SETools bug)".
                               format(rule.ruletype))

    def _reset_diff(self) -> None:
        """Reset diff results on policy changes."""
        self.log.debug("Resetting all *bounds differences")
        self.added_typebounds = None
        self.removed_typebounds = None

        # Sets of rules for each policy
        self._left_typebounds = None
        self._right_typebounds = None


class BoundsWrapper(Wrapper[Bounds]):

    """Wrap *bounds for diff purposes."""

    __slots__ = ("ruletype", "parent", "child")

    def __init__(self, rule: Bounds) -> None:
        self.origin = rule
        self.ruletype = rule.ruletype
        self.parent = type_wrapper_factory(rule.parent)
        self.child = type_wrapper_factory(rule.child)
        self.key = hash(rule)

    def __hash__(self):
        return self.key

    def __lt__(self, other):
        return self.key < other.key

    def __eq__(self, other):
        return self.ruletype == other.ruletype and \
            self.child == other.child



# Copyright 2016, Tresys Technology, LLC
# Copyright 2017, Chris PeBenito <pebenito@ieee.org>
#
# SPDX-License-Identifier: LGPL-2.1-only
#
from typing import NamedTuple

from ..policyrep import Context, Nodecon

from .context import ContextWrapper
from .descriptors import DiffResultDescriptor
from .difference import Difference, Wrapper


class ModifiedNodecon(NamedTuple):

    """Difference details for a modified netifcon."""

    rule: Nodecon
    added_context: Context
    removed_context: Context


class NodeconsDifference(Difference):

    """Determine the difference in nodecons between two policies."""

    added_nodecons = DiffResultDescriptor("diff_nodecons")
    removed_nodecons = DiffResultDescriptor("diff_nodecons")
    modified_nodecons = DiffResultDescriptor("diff_nodecons")

    def diff_nodecons(self) -> None:
        """Generate the difference in nodecons between the policies."""

        self.log.info("Generating nodecon differences from {0.left_policy} to {0.right_policy}".
                      format(self))

        self.added_nodecons, self.removed_nodecons, matched_nodecons = self._set_diff(
            (NodeconWrapper(n) for n in self.left_policy.nodecons()),
            (NodeconWrapper(n) for n in self.right_policy.nodecons()))

        self.modified_nodecons = []

        for left_nodecon, right_nodecon in matched_nodecons:
            # Criteria for modified nodecons
            # 1. change to context
            if ContextWrapper(left_nodecon.context) != ContextWrapper(right_nodecon.context):
                self.modified_nodecons.append(ModifiedNodecon(left_nodecon,
                                                              right_nodecon.context,
                                                              left_nodecon.context))

    #
    # Internal functions
    #
    def _reset_diff(self) -> None:
        """Reset diff results on policy changes."""
        self.log.debug("Resetting nodecon differences")
        self.added_nodecons = None
        self.removed_nodecons = None
        self.modified_nodecons = None


class NodeconWrapper(Wrapper[Nodecon]):

    """Wrap nodecon statements for diff purposes."""

    __slots__ = ("ip_version", "network")

    def __init__(self, ocon: Nodecon) -> None:
        self.origin = ocon
        self.ip_version = ocon.ip_version
        self.network = ocon.network
        self.key = hash(ocon)

    def __hash__(self):
        return self.key

    def __lt__(self, other):
        return self.origin < other.origin

    def __eq__(self, other):
        return self.ip_version == other.ip_version and \
            self.network == other.network



# Copyright 2014-2015, Tresys Technology, LLC
#
# SPDX-License-Identifier: LGPL-2.1-only
#
import logging
import re
from typing import cast, Iterable, Optional, Pattern, Union

from . import mixins, query
from .descriptors import CriteriaDescriptor, CriteriaSetDescriptor
from .exception import InvalidType, RuleUseError
from .policyrep import AnyRBACRule, RBACRuletype, Role, TypeOrAttr
from .util import match_indirect_regex


class RBACRuleQuery(mixins.MatchObjClass, query.PolicyQuery):

    """
    Query the RBAC rules.

    Parameter:
    policy            The policy to query.

    Keyword Parameters/Class attributes:
    ruletype        The list of rule type(s) to match.
    source          The name of the source role/attribute to match.
    source_indirect If true, members of an attribute will be
                    matched rather than the attribute itself.
    source_regex    If true, regular expression matching will
                    be used on the source role/attribute.
                    Obeys the source_indirect option.
    target          The name of the target role/attribute to match.
    target_indirect If true, members of an attribute will be
                    matched rather than the attribute itself.
    target_regex    If true, regular expression matching will
                    be used on the target role/attribute.
                    Obeys target_indirect option.
    tclass          The object class(es) to match.
    tclass_regex    If true, use a regular expression for
                    matching the rule's object class.
    default         The name of the default role to match.
    default_regex   If true, regular expression matching will
                    be used on the default role.
    """

    ruletype = CriteriaSetDescriptor(enum_class=RBACRuletype)
    source = CriteriaDescriptor("source_regex", "lookup_role")
    source_regex: bool = False
    source_indirect: bool = True
    _target: Optional[Union[Pattern, Role, TypeOrAttr]] = None
    target_regex: bool = False
    target_indirect: bool = True
    tclass = CriteriaSetDescriptor("tclass_regex", "lookup_class")
    tclass_regex: bool = False
    default = CriteriaDescriptor("default_regex", "lookup_role")
    default_regex: bool = False

    @property
    def target(self) -> Optional[Union[Pattern, Role, TypeOrAttr]]:
        return self._target

    @target.setter
    def target(self, value: Optional[Union[str, Role, TypeOrAttr]]) -> None:
        if not value:
            self._target = None
        elif self.target_regex:
            self._target = re.compile(value)
        else:
            try:
                self._target = self.policy.lookup_type_or_attr(cast(Union[str, TypeOrAttr], value))
            except InvalidType:
                self._target = self.policy.lookup_role(cast(Union[str, Role], value))

    def __init__(self, policy, **kwargs) -> None:
        super(RBACRuleQuery, self).__init__(policy, **kwargs)
        self.log = logging.getLogger(__name__)

    def results(self) -> Iterable[AnyRBACRule]:
        """Generator which yields all matching RBAC rules."""
        self.log.info("Generating RBAC rule results from {0.policy}".format(self))
        self.log.debug("Ruletypes: {0.ruletype}".format(self))
        self.log.debug("Source: {0.source!r}, indirect: {0.source_indirect}, "
                       "regex: {0.source_regex}".format(self))
        self.log.debug("Target: {0.target!r}, indirect: {0.target_indirect}, "
                       "regex: {0.target_regex}".format(self))
        self._match_object_class_debug(self.log)
        self.log.debug("Default: {0.default!r}, regex: {0.default_regex}".format(self))

        for rule in self.policy.rbacrules():
            #
            # Matching on rule type
            #
            if self.ruletype:
                if rule.ruletype not in self.ruletype:
                    continue

            #
            # Matching on source role
            #
            if self.source and not match_indirect_regex(
                    rule.source,
                    self.source,
                    self.source_indirect,
                    self.source_regex):
                continue

            #
            # Matching on target type (role_transition)/role(allow)
            #
            if self.target and not match_indirect_regex(
                    rule.target,
                    self.target,
                    self.target_indirect,
                    self.target_regex):
                continue

            #
            # Matching on object class
            #
            try:
                if not self._match_object_class(rule):
                    continue
            except RuleUseError:
                continue

            #
            # Matching on default role
            #
            if self.default:
                try:
                    # because default role is always a single
                    # role, hard-code indirect to True
                    # so the criteria can be an attribute
                    if not match_indirect_regex(
                            rule.default,
                            self.default,
                            True,
                            self.default_regex):
                        continue
                except RuleUseError:
                    continue

            # if we get here, we have matched all available criteria
            yield rule



# -*- Mode: Python; py-indent-offset: 4 -*-
# vim: tabstop=4 shiftwidth=4 expandtab
#
# Copyright (C) 2005-2009 Johan Dahlin <johan@gnome.org>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
# USA

# support overrides in different directories than our gi module
from pkgutil import extend_path
__path__ = extend_path(__path__, __name__)

import sys
import os
import importlib
import types

_static_binding_error = ('When using gi.repository you must not import static '
                         'modules like "gobject". Please change all occurrences '
                         'of "import gobject" to "from gi.repository import GObject". '
                         'See: https://bugzilla.gnome.org/show_bug.cgi?id=709183')

# we can't have pygobject 2 loaded at the same time we load the internal _gobject
if 'gobject' in sys.modules:
    raise ImportError(_static_binding_error)


from . import _gi
from ._gi import _API
from ._gi import Repository
from ._gi import PyGIDeprecationWarning
from ._gi import PyGIWarning

_API = _API  # pyflakes
PyGIDeprecationWarning = PyGIDeprecationWarning
PyGIWarning = PyGIWarning

_versions = {}
_overridesdir = os.path.join(os.path.dirname(__file__), 'overrides')

# Needed for compatibility with "pygobject.h"/pygobject_init()
_gobject = types.ModuleType("gi._gobject")
sys.modules[_gobject.__name__] = _gobject
_gobject._PyGObject_API = _gi._PyGObject_API
_gobject.pygobject_version = _gi.pygobject_version

version_info = _gi.pygobject_version[:]
__version__ = "{0}.{1}.{2}".format(*version_info)

_gi.register_foreign()


class _DummyStaticModule(types.ModuleType):
    __path__ = None

    def __getattr__(self, name):
        raise AttributeError(_static_binding_error)


sys.modules['glib'] = _DummyStaticModule('glib', _static_binding_error)
sys.modules['gobject'] = _DummyStaticModule('gobject', _static_binding_error)
sys.modules['gio'] = _DummyStaticModule('gio', _static_binding_error)
sys.modules['gtk'] = _DummyStaticModule('gtk', _static_binding_error)
sys.modules['gtk.gdk'] = _DummyStaticModule('gtk.gdk', _static_binding_error)


def check_version(version):
    if isinstance(version, str):
        version_list = tuple(map(int, version.split(".")))
    else:
        version_list = version

    if version_list > version_info:
        raise ValueError((
            "pygobject's version %s required, and available version "
            "%s is not recent enough") % (version, __version__)
        )


def require_version(namespace, version):
    """ Ensures the correct versions are loaded when importing `gi` modules.

    :param namespace: The name of module to require.
    :type namespace: str
    :param version: The version of module to require.
    :type version: str
    :raises ValueError: If module/version is already loaded, already required, or unavailable.

    :Example:

    .. code-block:: python

        import gi
        gi.require_version('Gtk', '3.0')

    """
    repository = Repository.get_default()

    if not isinstance(version, str):
        raise ValueError('Namespace version needs to be a string.')

    if namespace in repository.get_loaded_namespaces():
        loaded_version = repository.get_version(namespace)
        if loaded_version != version:
            raise ValueError('Namespace %s is already loaded with version %s' %
                             (namespace, loaded_version))

    if namespace in _versions and _versions[namespace] != version:
        raise ValueError('Namespace %s already requires version %s' %
                         (namespace, _versions[namespace]))

    available_versions = repository.enumerate_versions(namespace)
    if not available_versions:
        raise ValueError('Namespace %s not available' % namespace)

    if version not in available_versions:
        raise ValueError('Namespace %s not available for version %s' %
                         (namespace, version))

    _versions[namespace] = version


def require_versions(requires):
    """ Utility function for consolidating multiple `gi.require_version()` calls.

    :param requires: The names and versions of modules to require.
    :type requires: dict

    :Example:

    .. code-block:: python

        import gi
        gi.require_versions({'Gtk': '3.0', 'GLib': '2.0', 'Gio': '2.0'})
    """
    for module_name, module_version in requires.items():
        require_version(module_name, module_version)


def get_required_version(namespace):
    return _versions.get(namespace, None)


def require_foreign(namespace, symbol=None):
    """Ensure the given foreign marshaling module is available and loaded.

    :param str namespace:
        Introspection namespace of the foreign module (e.g. "cairo")
    :param symbol:
        Optional symbol typename to ensure a converter exists.
    :type symbol: str or None
    :raises: ImportError

    :Example:

    .. code-block:: python

        import gi
        import cairo
        gi.require_foreign('cairo')

    """
    try:
        _gi.require_foreign(namespace, symbol)
    except Exception as e:
        raise ImportError(str(e))
    importlib.import_module('gi.repository', namespace)



# -*- coding: utf-8 -*-

from __future__ import absolute_import, print_function, unicode_literals

import re
import os
import warnings
import weakref

from . import gpgme
from .errors import errorcheck, GPGMEError
from . import constants
from . import errors
from . import util

del absolute_import, print_function, unicode_literals

# Copyright (C) 2016-2018 g10 Code GmbH
# Copyright (C) 2004, 2008 Igor Belyi <belyi@users.sourceforge.net>
# Copyright (C) 2002 John Goerzen <jgoerzen@complete.org>
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
"""Core functionality

Core functionality of GPGME wrapped in a object-oriented fashion.
Provides the 'Context' class for performing cryptographic operations,
and the 'Data' class describing buffers of data.

"""


class GpgmeWrapper(object):
    """Base wrapper class

    Not to be instantiated directly.

    """

    def __init__(self, wrapped):
        self._callback_excinfo = None
        self.wrapped = wrapped

    def __repr__(self):
        return '<{}/{!r}>'.format(
            super(GpgmeWrapper, self).__repr__(), self.wrapped)

    def __str__(self):
        acc = ['{}.{}'.format(__name__, self.__class__.__name__)]
        flags = [f for f in self._boolean_properties if getattr(self, f)]
        if flags:
            acc.append('({})'.format(' '.join(flags)))

        return '<{}>'.format(' '.join(acc))

    def __hash__(self):
        return hash(repr(self.wrapped))

    def __eq__(self, other):
        if other is None:
            return False
        else:
            return repr(self.wrapped) == repr(other.wrapped)

    @property
    def _ctype(self):
        """The name of the c type wrapped by this class

        Must be set by child classes.

        """
        raise NotImplementedError()

    @property
    def _cprefix(self):
        """The common prefix of c functions wrapped by this class

        Must be set by child classes.

        """
        raise NotImplementedError()

    def _errorcheck(self, name):
        """Must be implemented by child classes.

        This function must return a trueish value for all c functions
        returning gpgme_error_t."""
        raise NotImplementedError()

    """The set of all boolean properties"""
    _boolean_properties = set()

    def __wrap_boolean_property(self, key, do_set=False, value=None):
        get_func = getattr(gpgme, "{}get_{}".format(self._cprefix, key))
        set_func = getattr(gpgme, "{}set_{}".format(self._cprefix, key))

        def get(slf):
            return bool(get_func(slf.wrapped))

        def set_(slf, value):
            set_func(slf.wrapped, bool(value))

        p = property(get, set_, doc="{} flag".format(key))
        setattr(self.__class__, key, p)

        if do_set:
            set_(self, bool(value))
        else:
            return get(self)

    _munge_docstring = re.compile(r'gpgme_([^(]*)\(([^,]*), (.*\) -> .*)')

    def __getattr__(self, key):
        """On-the-fly generation of wrapper methods and properties"""
        if key[0] == '_' or self._cprefix is None:
            return None

        if key in self._boolean_properties:
            return self.__wrap_boolean_property(key)

        name = self._cprefix + key
        func = getattr(gpgme, name)

        if self._errorcheck(name):

            def _funcwrap(slf, *args):
                result = func(slf.wrapped, *args)
                if slf._callback_excinfo:
                    gpgme.gpg_raise_callback_exception(slf)
                return errorcheck(result, name)
        else:

            def _funcwrap(slf, *args):
                result = func(slf.wrapped, *args)
                if slf._callback_excinfo:
                    gpgme.gpg_raise_callback_exception(slf)
                return result

        doc_orig = getattr(func, "__doc__")
        if doc_orig:
            doc = self._munge_docstring.sub(r'\2.\1(\3', doc_orig)
        else:
            doc = None

        _funcwrap.__doc__ = doc

        # Monkey-patch the class.
        setattr(self.__class__, key, _funcwrap)

        # Bind the method to 'self'.
        def wrapper(*args):
            return _funcwrap(self, *args)

        wrapper.__doc__ = doc

        return wrapper

    def __setattr__(self, key, value):
        """On-the-fly generation of properties"""
        if key in self._boolean_properties:
            self.__wrap_boolean_property(key, True, value)
        else:
            super(GpgmeWrapper, self).__setattr__(key, value)


class Context(GpgmeWrapper):
    """Context for cryptographic operations

    All cryptographic operations in GPGME are performed within a
    context, which contains the internal state of the operation as
    well as configuration parameters.  By using several contexts you
    can run several cryptographic operations in parallel, with
    different configuration.

    Access to a context must be synchronized.

    """

    def __init__(self,
                 armor=False,
                 textmode=False,
                 offline=False,
                 signers=[],
                 pinentry_mode=constants.PINENTRY_MODE_DEFAULT,
                 protocol=constants.PROTOCOL_OpenPGP,
                 wrapped=None,
                 home_dir=None):
        """Construct a context object

        Keyword arguments:
        armor		-- enable ASCII armoring (default False)
        textmode	-- enable canonical text mode (default False)
        offline		-- do not contact external key sources (default False)
        signers		-- list of keys used for signing (default [])
        pinentry_mode	-- pinentry mode (default PINENTRY_MODE_DEFAULT)
        protocol	-- protocol to use (default PROTOCOL_OpenPGP)
        home_dir        -- state directory (default is the engine default)

        """
        if wrapped:
            self.own = False
        else:
            tmp = gpgme.new_gpgme_ctx_t_p()
            errorcheck(gpgme.gpgme_new(tmp))
            wrapped = gpgme.gpgme_ctx_t_p_value(tmp)
            gpgme.delete_gpgme_ctx_t_p(tmp)
            self.own = True
        super(Context, self).__init__(wrapped)
        self.armor = armor
        self.textmode = textmode
        self.offline = offline
        self.signers = signers
        self.pinentry_mode = pinentry_mode
        self.protocol = protocol
        self.home_dir = home_dir

    def __read__(self, sink, data):
        """Read helper

        Helper function to retrieve the results of an operation, or
        None if SINK is given.
        """
        if sink or data is None:
            return None
        data.seek(0, os.SEEK_SET)
        return data.read()

    def __repr__(self):
        return ("Context(armor={0.armor}, "
                "textmode={0.textmode}, offline={0.offline}, "
                "signers={0.signers}, pinentry_mode={0.pinentry_mode}, "
                "protocol={0.protocol}, home_dir={0.home_dir}"
                ")").format(self)

    def encrypt(self,
                plaintext,
                recipients=[],
                sign=True,
                sink=None,
                passphrase=None,
                always_trust=False,
                add_encrypt_to=False,
                prepare=False,
                expect_sign=False,
                compress=True):
        """Encrypt data

        Encrypt the given plaintext for the given recipients.  If the
        list of recipients is empty, the data is encrypted
        symmetrically with a passphrase.

        The passphrase can be given as parameter, using a callback
        registered at the context, or out-of-band via pinentry.

        Keyword arguments:
        recipients	-- list of keys to encrypt to
        sign		-- sign plaintext (default True)
        sink		-- write result to sink instead of returning it
        passphrase	-- for symmetric encryption
        always_trust	-- always trust the keys (default False)
        add_encrypt_to	-- encrypt to configured additional keys (default False)
        prepare		-- (ui) prepare for encryption (default False)
        expect_sign	-- (ui) prepare for signing (default False)
        compress	-- compress plaintext (default True)

        Returns:
        ciphertext	-- the encrypted data (or None if sink is given)
        result		-- additional information about the encryption
        sign_result	-- additional information about the signature(s)

        Raises:
        InvalidRecipients -- if encryption using a particular key failed
        InvalidSigners	-- if signing using a particular key failed
        GPGMEError	-- as signaled by the underlying library

        """
        ciphertext = sink if sink else Data()
        flags = 0
        flags |= always_trust * constants.ENCRYPT_ALWAYS_TRUST
        flags |= (not add_encrypt_to) * constants.ENCRYPT_NO_ENCRYPT_TO
        flags |= prepare * constants.ENCRYPT_PREPARE
        flags |= expect_sign * constants.ENCRYPT_EXPECT_SIGN
        flags |= (not compress) * constants.ENCRYPT_NO_COMPRESS

        if passphrase is not None:
            old_pinentry_mode = self.pinentry_mode
            old_passphrase_cb = getattr(self, '_passphrase_cb', None)
            self.pinentry_mode = constants.PINENTRY_MODE_LOOPBACK

            def passphrase_cb(hint, desc, prev_bad, hook=None):
                return passphrase

            self.set_passphrase_cb(passphrase_cb)

        try:
            if sign:
                self.op_encrypt_sign(recipients, flags, plaintext, ciphertext)
            else:
                self.op_encrypt(recipients, flags, plaintext, ciphertext)
        except errors.GPGMEError as e:
            result = self.op_encrypt_result()
            sig_result = self.op_sign_result() if sign else None
            results = (self.__read__(sink, ciphertext), result, sig_result)
            if e.getcode() == errors.UNUSABLE_PUBKEY:
                if result.invalid_recipients:
                    raise errors.InvalidRecipients(
                        result.invalid_recipients,
                        error=e.error,
                        results=results)
            if e.getcode() == errors.UNUSABLE_SECKEY:
                sig_result = self.op_sign_result()
                if sig_result.invalid_signers:
                    raise errors.InvalidSigners(
                        sig_result.invalid_signers,
                        error=e.error,
                        results=results)
            # Otherwise, just raise the error, but attach the results
            # first.
            e.results = results
            raise e
        finally:
            if passphrase is not None:
                self.pinentry_mode = old_pinentry_mode
                if old_passphrase_cb:
                    self.set_passphrase_cb(*old_passphrase_cb[1:])

        result = self.op_encrypt_result()
        assert not result.invalid_recipients
        sig_result = self.op_sign_result() if sign else None
        assert not sig_result or not sig_result.invalid_signers

        return self.__read__(sink, ciphertext), result, sig_result

    def decrypt(self, ciphertext, sink=None, passphrase=None, verify=True):
        """Decrypt data

        Decrypt the given ciphertext and verify any signatures.  If
        VERIFY is an iterable of keys, the ciphertext must be signed
        by all those keys, otherwise a MissingSignatures error is
        raised.  Note: if VERIFY is an empty iterable, that is treated
        the same as passing verify=True (that is, verify signatures
        and return data about any valid signatures found, but no
        signatures are required and no MissingSignatures error will be
        raised).

        If the ciphertext is symmetrically encrypted using a
        passphrase, that passphrase can be given as parameter, using a
        callback registered at the context, or out-of-band via
        pinentry.

        Keyword arguments:
        sink            -- write result to sink instead of returning it
        passphrase      -- for symmetric decryption
        verify          -- check signatures (boolean or iterable of keys,
                           see above) (default True)

        Returns:
        plaintext       -- the decrypted data (or None if sink is given)
        result          -- additional information about the decryption
        verify_result   -- additional information about the valid
                           signature(s) found

        Raises:
        UnsupportedAlgorithm -- if an unsupported algorithm was used
        MissingSignatures    -- if expected signatures are missing or bad
        GPGMEError           -- as signaled by the underlying library

        """
        do_sig_verification = False
        required_keys = None
        plaintext = sink if sink else Data()

        if passphrase is not None:
            old_pinentry_mode = self.pinentry_mode
            old_passphrase_cb = getattr(self, '_passphrase_cb', None)
            self.pinentry_mode = constants.PINENTRY_MODE_LOOPBACK

            def passphrase_cb(hint, desc, prev_bad, hook=None):
                return passphrase

            self.set_passphrase_cb(passphrase_cb)

        try:
            if isinstance(verify, bool):
                do_sig_verification = verify
            elif verify is None:
                warnings.warn(
                    "ctx.decrypt called with verify=None, should be bool or iterable (treating as False).",
                    category=DeprecationWarning)
                do_sig_verification = False
            else:
                # we hope this is an iterable:
                required_keys = verify
                do_sig_verification = True

            if do_sig_verification:
                self.op_decrypt_verify(ciphertext, plaintext)
            else:
                self.op_decrypt(ciphertext, plaintext)
        except errors.GPGMEError as e:
            result = self.op_decrypt_result()
            if do_sig_verification:
                verify_result = self.op_verify_result()
            else:
                verify_result = None
            # Just raise the error, but attach the results first.
            e.results = (self.__read__(sink, plaintext), result, verify_result)
            raise e
        finally:
            if passphrase is not None:
                self.pinentry_mode = old_pinentry_mode
                if old_passphrase_cb:
                    self.set_passphrase_cb(*old_passphrase_cb[1:])

        result = self.op_decrypt_result()

        if do_sig_verification:
            verify_result = self.op_verify_result()
        else:
            verify_result = None

        results = (self.__read__(sink, plaintext), result, verify_result)

        if result.unsupported_algorithm:
            raise errors.UnsupportedAlgorithm(result.unsupported_algorithm,
                                              results=results)

        if do_sig_verification:
            # filter out all invalid signatures
            verify_result.signatures = list(filter(lambda s: s.status == errors.NO_ERROR, verify_result.signatures))
            if required_keys is not None:
                missing = []
                for key in required_keys:
                    ok = False
                    for subkey in key.subkeys:
                        for sig in verify_result.signatures:
                            if sig.summary & constants.SIGSUM_VALID == 0:
                                continue
                            if subkey.can_sign and subkey.fpr == sig.fpr:
                                ok = True
                            break
                        if ok:
                            break
                    if not ok:
                        missing.append(key)
                if missing:
                    raise errors.MissingSignatures(verify_result, missing,
                                                   results=results)

        return results

    def sign(self, data, sink=None, mode=constants.SIG_MODE_NORMAL):
        """Sign data

        Sign the given data with either the configured default local
        key, or the 'signers' keys of this context.

        Keyword arguments:
        mode		-- signature mode (default: normal, see below)
        sink		-- write result to sink instead of returning it

        Returns:
        either
          signed_data	-- encoded data and signature (normal mode)
          signature	-- only the signature data (detached mode)
          cleartext	-- data and signature as text (cleartext mode)
            (or None if sink is given)
        result		-- additional information about the signature(s)

        Raises:
        InvalidSigners	-- if signing using a particular key failed
        GPGMEError	-- as signaled by the underlying library

        """
        signeddata = sink if sink else Data()

        try:
            self.op_sign(data, signeddata, mode)
        except errors.GPGMEError as e:
            results = (self.__read__(sink, signeddata), self.op_sign_result())
            if e.getcode() == errors.UNUSABLE_SECKEY:
                if results[1].invalid_signers:
                    raise errors.InvalidSigners(
                        results[1].invalid_signers,
                        error=e.error,
                        results=results)
            e.results = results
            raise e

        result = self.op_sign_result()
        assert not result.invalid_signers

        return self.__read__(sink, signeddata), result

    def verify(self, signed_data, signature=None, sink=None, verify=[]):
        """Verify signatures

        Verify signatures over data.  If VERIFY is an iterable of
        keys, the ciphertext must be signed by all those keys,
        otherwise an error is raised.

        Keyword arguments:
        signature	-- detached signature data
        sink		-- write result to sink instead of returning it

        Returns:
        data		-- the plain data
            (or None if sink is given, or we verified a detached signature)
        result		-- additional information about the signature(s)

        Raises:
        BadSignatures	-- if a bad signature is encountered
        MissingSignatures -- if expected signatures are missing or bad
        GPGMEError	-- as signaled by the underlying library

        """
        if signature:
            # Detached signature, we don't return the plain text.
            data = None
        else:
            data = sink if sink else Data()

        try:
            if signature:
                self.op_verify(signature, signed_data, None)
            else:
                self.op_verify(signed_data, None, data)
        except errors.GPGMEError as e:
            # Just raise the error, but attach the results first.
            e.results = (self.__read__(sink, data), self.op_verify_result())
            raise e

        results = (self.__read__(sink, data), self.op_verify_result())
        if any(s.status != errors.NO_ERROR for s in results[1].signatures):
            raise errors.BadSignatures(results[1], results=results)

        missing = list()
        for key in verify:
            ok = False
            for subkey in key.subkeys:
                for sig in results[1].signatures:
                    if sig.summary & constants.SIGSUM_VALID == 0:
                        continue
                    if subkey.can_sign and subkey.fpr == sig.fpr:
                        ok = True
                        break
                if ok:
                    break
            if not ok:
                missing.append(key)
        if missing:
            raise errors.MissingSignatures(
                results[1], missing, results=results)

        return results

    def key_import(self, data):
        """Import data

        Imports the given data into the Context.

        Returns:
                -- an object describing the results of imported or updated
                   keys

        Raises:
        TypeError      -- Very rarely.
        GPGMEError     -- as signaled by the underlying library:

                          Import status errors, when they occur, will usually
                          be of NODATA.  NO_PUBKEY indicates something
                          managed to run the function without any
                          arguments, while an argument of None triggers
                          the first NODATA of errors.GPGME in the
                          exception.
        """
        try:
            self.op_import(data)
            result = self.op_import_result()
            if result.considered == 0:
                status = constants.STATUS_IMPORT_PROBLEM
            else:
                status = constants.STATUS_KEY_CONSIDERED
        except Exception as e:
            if e == errors.GPGMEError:
                if e.code_str == "No data":
                    status = constants.STATUS_NODATA
                else:
                    status = constants.STATUS_FILE_ERROR
            elif e == TypeError and hasattr(data, "decode") is True:
                status = constants.STATUS_NO_PUBKEY
            elif e == TypeError and hasattr(data, "encode") is True:
                status = constants.STATUS_FILE_ERROR
            else:
                status = constants.STATUS_ERROR

        if status == constants.STATUS_KEY_CONSIDERED:
            import_result = result
        else:
            import_result = status

        return import_result

    def key_export(self, pattern=None):
        """Export keys.

        Exports public keys matching the pattern specified.  If no
        pattern is specified then exports all available keys.

        Keyword arguments:
        pattern	-- return keys matching pattern (default: all keys)

        Returns:
                -- A key block containing one or more OpenPGP keys in
                   either ASCII armoured or binary format as determined
                   by the Context().  If there are no matching keys it
                   returns None.

        Raises:
        GPGMEError     -- as signaled by the underlying library.
        """
        data = Data()
        mode = 0
        try:
            self.op_export(pattern, mode, data)
            data.seek(0, os.SEEK_SET)
            pk_result = data.read()
        except GPGMEError as e:
            raise e

        if len(pk_result) > 0:
            result = pk_result
        else:
            result = None

        return result

    def key_export_minimal(self, pattern=None):
        """Export keys.

        Exports public keys matching the pattern specified in a
        minimised format.  If no pattern is specified then exports all
        available keys.

        Keyword arguments:
        pattern	-- return keys matching pattern (default: all keys)

        Returns:
                -- A key block containing one or more minimised OpenPGP
                   keys in either ASCII armoured or binary format as
                   determined by the Context().  If there are no matching
                   keys it returns None.

        Raises:
        GPGMEError     -- as signaled by the underlying library.
        """
        data = Data()
        mode = gpgme.GPGME_EXPORT_MODE_MINIMAL
        try:
            self.op_export(pattern, mode, data)
            data.seek(0, os.SEEK_SET)
            pk_result = data.read()
        except GPGMEError as e:
            raise e

        if len(pk_result) > 0:
            result = pk_result
        else:
            result = None

        return result

    def key_export_secret(self, pattern=None):
        """Export secret keys.

        Exports secret keys matching the pattern specified.  If no
        pattern is specified then exports or attempts to export all
        available secret keys.

        IMPORTANT: Each secret key to be exported will prompt for its
        passphrase via an invocation of pinentry and gpg-agent.  If the
        passphrase is not entered or does not match then no data will be
        exported.  This is the same result as when specifying a pattern
        that is not matched by the available keys.

        Keyword arguments:
        pattern	-- return keys matching pattern (default: all keys)

        Returns:
                -- On success a key block containing one or more OpenPGP
                   secret keys in either ASCII armoured or binary format
                   as determined by the Context().
                -- On failure while not raising an exception, returns None.

        Raises:
        GPGMEError     -- as signaled by the underlying library.
        """
        data = Data()
        mode = gpgme.GPGME_EXPORT_MODE_SECRET
        try:
            self.op_export(pattern, mode, data)
            data.seek(0, os.SEEK_SET)
            sk_result = data.read()
        except GPGMEError as e:
            raise e

        if len(sk_result) > 0:
            result = sk_result
        else:
            result = None

        return result

    def keylist(self,
                pattern=None,
                secret=False,
                mode=constants.keylist.mode.LOCAL,
                source=None):
        """List keys

        Keyword arguments:
        pattern	-- return keys matching pattern (default: all keys)
        secret	-- return only secret keys (default: False)
        mode    -- keylist mode (default: list local keys)
        source  -- read keys from source instead from the keyring
                       (all other options are ignored in this case)

        Returns:
                -- an iterator returning key objects

        Raises:
        GPGMEError	-- as signaled by the underlying library
        """
        if not source:
            self.set_keylist_mode(mode)
            self.op_keylist_start(pattern, secret)
        else:
            # Automatic wrapping of SOURCE is not possible here,
            # because the object must not be deallocated until the
            # iteration over the results ends.
            if not isinstance(source, Data):
                source = Data(file=source)
            self.op_keylist_from_data_start(source, 0)

        key = self.op_keylist_next()
        while key:
            yield key
            key = self.op_keylist_next()
        self.op_keylist_end()

    def create_key(self,
                   userid,
                   algorithm=None,
                   expires_in=0,
                   expires=True,
                   sign=False,
                   encrypt=False,
                   certify=False,
                   authenticate=False,
                   passphrase=None,
                   force=False):
        """Create a primary key

        Create a primary key for the user id USERID.

        ALGORITHM may be used to specify the public key encryption
        algorithm for the new key.  By default, a reasonable default
        is chosen.  You may use "future-default" to select an
        algorithm that will be the default in a future implementation
        of the engine.  ALGORITHM may be a string like "rsa", or
        "rsa2048" to explicitly request an algorithm and a key size.

        EXPIRES_IN specifies the expiration time of the key in number
        of seconds since the keys creation.  By default, a reasonable
        expiration time is chosen.  If you want to create a key that
        does not expire, use the keyword argument EXPIRES.

        SIGN, ENCRYPT, CERTIFY, and AUTHENTICATE can be used to
        request the capabilities of the new key.  If you don't request
        any, a reasonable set of capabilities is selected, and in case
        of OpenPGP, a subkey with a reasonable set of capabilities is
        created.

        If PASSPHRASE is None (the default), then the key will not be
        protected with a passphrase.  If PASSPHRASE is a string, it
        will be used to protect the key.  If PASSPHRASE is True, the
        passphrase must be supplied using a passphrase callback or
        out-of-band with a pinentry.

        Keyword arguments:
        algorithm    -- public key algorithm, see above (default: reasonable)
        expires_in   -- expiration time in seconds (default: reasonable)
        expires      -- whether or not the key should expire (default: True)
        sign         -- request the signing capability (see above)
        encrypt      -- request the encryption capability (see above)
        certify      -- request the certification capability (see above)
        authenticate -- request the authentication capability (see above)
        passphrase   -- protect the key with a passphrase (default: no
                        passphrase)
        force        -- force key creation even if a key with the same userid
                        exists (default: False)

        Returns:
                     -- an object describing the result of the key creation

        Raises:
        GPGMEError   -- as signaled by the underlying library

        """
        if util.is_a_string(passphrase):
            old_pinentry_mode = self.pinentry_mode
            old_passphrase_cb = getattr(self, '_passphrase_cb', None)
            self.pinentry_mode = constants.PINENTRY_MODE_LOOPBACK

            def passphrase_cb(hint, desc, prev_bad, hook=None):
                return passphrase

            self.set_passphrase_cb(passphrase_cb)

        try:
            self.op_createkey(
                userid,
                algorithm,
                0,  # reserved
                expires_in,
                None,  # extrakey
                ((constants.create.SIGN if sign else 0) |
                 (constants.create.ENCR if encrypt else 0) |
                 (constants.create.CERT if certify else 0) |
                 (constants.create.AUTH if authenticate else 0) |
                 (constants.create.NOPASSWD if passphrase is None else 0) |
                 (0 if expires else constants.create.NOEXPIRE) |
                 (constants.create.FORCE if force else 0)))
        finally:
            if util.is_a_string(passphrase):
                self.pinentry_mode = old_pinentry_mode
                if old_passphrase_cb:
                    self.set_passphrase_cb(*old_passphrase_cb[1:])

        return self.op_genkey_result()

    def create_subkey(self,
                      key,
                      algorithm=None,
                      expires_in=0,
                      expires=True,
                      sign=False,
                      encrypt=False,
                      authenticate=False,
                      passphrase=None):
        """Create a subkey

        Create a subkey for the given KEY.  As subkeys are a concept
        of OpenPGP, calling this is only valid for the OpenPGP
        protocol.

        ALGORITHM may be used to specify the public key encryption
        algorithm for the new subkey.  By default, a reasonable
        default is chosen.  You may use "future-default" to select an
        algorithm that will be the default in a future implementation
        of the engine.  ALGORITHM may be a string like "rsa", or
        "rsa2048" to explicitly request an algorithm and a key size.

        EXPIRES_IN specifies the expiration time of the subkey in
        number of seconds since the subkeys creation.  By default, a
        reasonable expiration time is chosen.  If you want to create a
        subkey that does not expire, use the keyword argument EXPIRES.

        SIGN, ENCRYPT, and AUTHENTICATE can be used to request the
        capabilities of the new subkey.  If you don't request any, an
        encryption subkey is generated.

        If PASSPHRASE is None (the default), then the subkey will not
        be protected with a passphrase.  If PASSPHRASE is a string, it
        will be used to protect the subkey.  If PASSPHRASE is True,
        the passphrase must be supplied using a passphrase callback or
        out-of-band with a pinentry.

        Keyword arguments:
        algorithm    -- public key algorithm, see above (default: reasonable)
        expires_in   -- expiration time in seconds (default: reasonable)
        expires      -- whether or not the subkey should expire (default: True)
        sign         -- request the signing capability (see above)
        encrypt      -- request the encryption capability (see above)
        authenticate -- request the authentication capability (see above)
        passphrase   -- protect the subkey with a passphrase (default: no
                        passphrase)

        Returns:
                     -- an object describing the result of the subkey creation

        Raises:
        GPGMEError   -- as signaled by the underlying library

        """
        if util.is_a_string(passphrase):
            old_pinentry_mode = self.pinentry_mode
            old_passphrase_cb = getattr(self, '_passphrase_cb', None)
            self.pinentry_mode = constants.PINENTRY_MODE_LOOPBACK

            def passphrase_cb(hint, desc, prev_bad, hook=None):
                return passphrase

            self.set_passphrase_cb(passphrase_cb)

        try:
            self.op_createsubkey(
                key,
                algorithm,
                0,  # reserved
                expires_in,
                ((constants.create.SIGN if sign else 0) |
                 (constants.create.ENCR if encrypt else 0) |
                 (constants.create.AUTH if authenticate else 0) |
                 (constants.create.NOPASSWD if passphrase is None else 0) |
                 (0 if expires else constants.create.NOEXPIRE)))
        finally:
            if util.is_a_string(passphrase):
                self.pinentry_mode = old_pinentry_mode
                if old_passphrase_cb:
                    self.set_passphrase_cb(*old_passphrase_cb[1:])

        return self.op_genkey_result()

    def key_add_uid(self, key, uid):
        """Add a UID

        Add the uid UID to the given KEY.  Calling this function is
        only valid for the OpenPGP protocol.

        Raises:
        GPGMEError   -- as signaled by the underlying library

        """
        self.op_adduid(key, uid, 0)

    def key_revoke_uid(self, key, uid):
        """Revoke a UID

        Revoke the uid UID from the given KEY.  Calling this function
        is only valid for the OpenPGP protocol.

        Raises:
        GPGMEError   -- as signaled by the underlying library

        """
        self.op_revuid(key, uid, 0)

    def key_sign(self, key, uids=None, expires_in=False, local=False):
        """Sign a key

        Sign a key with the current set of signing keys.  Calling this
        function is only valid for the OpenPGP protocol.

        If UIDS is None (the default), then all UIDs are signed.  If
        it is a string, then only the matching UID is signed.  If it
        is a list of strings, then all matching UIDs are signed.  Note
        that a case-sensitive exact string comparison is done.

        EXPIRES_IN specifies the expiration time of the signature in
        seconds.  If EXPIRES_IN is False, the signature does not
        expire.

        Keyword arguments:
        uids         -- user ids to sign, see above (default: sign all)
        expires_in   -- validity period of the signature in seconds
                                               (default: do not expire)
        local        -- create a local, non-exportable signature
                                               (default: False)

        Raises:
        GPGMEError   -- as signaled by the underlying library

        """
        flags = 0
        if uids is None or util.is_a_string(uids):
            pass  # through unchanged
        else:
            flags |= constants.keysign.LFSEP
            uids = "\n".join(uids)

        if not expires_in:
            flags |= constants.keysign.NOEXPIRE

        if local:
            flags |= constants.keysign.LOCAL

        self.op_keysign(key, uids, expires_in, flags)

    def key_tofu_policy(self, key, policy):
        """Set a keys' TOFU policy

        Set the TOFU policy associated with KEY to POLICY.  Calling
        this function is only valid for the OpenPGP protocol.

        Raises:
        GPGMEError   -- as signaled by the underlying library

        """
        self.op_tofu_policy(key, policy)

    def assuan_transact(self,
                        command,
                        data_cb=None,
                        inquire_cb=None,
                        status_cb=None):
        """Issue a raw assuan command

        This function can be used to issue a raw assuan command to the
        engine.

        If command is a string or bytes, it will be used as-is.  If it
        is an iterable of strings, it will be properly escaped and
        joined into an well-formed assuan command.

        Keyword arguments:
        data_cb		-- a callback receiving data lines
        inquire_cb	-- a callback providing more information
        status_cb	-- a callback receiving status lines

        Returns:
        result		-- the result of command as GPGMEError

        Raises:
        GPGMEError	-- as signaled by the underlying library

        """

        if util.is_a_string(command) or isinstance(command, bytes):
            cmd = command
        else:
            cmd = " ".join(util.percent_escape(f) for f in command)

        errptr = gpgme.new_gpgme_error_t_p()

        err = gpgme.gpgme_op_assuan_transact_ext(
            self.wrapped, cmd, (weakref.ref(self), data_cb)
            if data_cb else None, (weakref.ref(self), inquire_cb)
            if inquire_cb else None, (weakref.ref(self), status_cb)
            if status_cb else None, errptr)

        if self._callback_excinfo:
            gpgme.gpg_raise_callback_exception(self)

        errorcheck(err)

        status = gpgme.gpgme_error_t_p_value(errptr)
        gpgme.delete_gpgme_error_t_p(errptr)

        return GPGMEError(status) if status != 0 else None

    def interact(self, key, func, sink=None, flags=0, fnc_value=None):
        """Interact with the engine

        This method can be used to edit keys and cards interactively.
        KEY is the key to edit, FUNC is called repeatedly with two
        unicode arguments, 'keyword' and 'args'.  See the GPGME manual
        for details.

        Keyword arguments:
        sink		-- if given, additional output is written here
        flags		-- use constants.INTERACT_CARD to edit a card

        Raises:
        GPGMEError	-- as signaled by the underlying library

        """
        if key is None:
            raise ValueError("First argument cannot be None")

        if sink is None:
            sink = Data()

        if fnc_value:
            opaquedata = (weakref.ref(self), func, fnc_value)
        else:
            opaquedata = (weakref.ref(self), func)

        result = gpgme.gpgme_op_interact(self.wrapped, key, flags, opaquedata,
                                         sink)
        if self._callback_excinfo:
            gpgme.gpg_raise_callback_exception(self)
        errorcheck(result)

    @property
    def signers(self):
        """Keys used for signing"""
        return [self.signers_enum(i) for i in range(self.signers_count())]

    @signers.setter
    def signers(self, signers):
        old = self.signers
        self.signers_clear()
        try:
            for key in signers:
                self.signers_add(key)
        except:
            self.signers = old
            raise

    @property
    def pinentry_mode(self):
        """Pinentry mode"""
        return self.get_pinentry_mode()

    @pinentry_mode.setter
    def pinentry_mode(self, value):
        self.set_pinentry_mode(value)

    @property
    def protocol(self):
        """Protocol to use"""
        return self.get_protocol()

    @protocol.setter
    def protocol(self, value):
        errorcheck(gpgme.gpgme_engine_check_version(value))
        self.set_protocol(value)

    @property
    def home_dir(self):
        """Engine's home directory"""
        return self.engine_info.home_dir

    @home_dir.setter
    def home_dir(self, value):
        self.set_engine_info(self.protocol, home_dir=value)

    _ctype = 'gpgme_ctx_t'
    _cprefix = 'gpgme_'

    def _errorcheck(self, name):
        """This function should list all functions returning gpgme_error_t"""
        # The list of functions is created using:
        #
        # $ grep '^gpgme_error_t ' obj/lang/python/python3.5-gpg/gpgme.h \
        # | grep -v _op_ | awk "/\(gpgme_ctx/ { printf (\"'%s',\\n\", \$2) } "
        return ((name.startswith('gpgme_op_') and not
                 name.endswith('_result')) or name in {
                     'gpgme_new', 'gpgme_set_ctx_flag', 'gpgme_set_protocol',
                     'gpgme_set_sub_protocol', 'gpgme_set_keylist_mode',
                     'gpgme_set_pinentry_mode', 'gpgme_set_locale',
                     'gpgme_ctx_set_engine_info', 'gpgme_signers_add',
                     'gpgme_sig_notation_add', 'gpgme_set_sender',
                     'gpgme_cancel', 'gpgme_cancel_async', 'gpgme_get_key',
                     'gpgme_get_sig_key',
                })

    _boolean_properties = {'armor', 'textmode', 'offline'}

    def __del__(self):
        if not gpgme:
            # At interpreter shutdown, gpgme is set to NONE.
            return

        self._free_passcb()
        self._free_progresscb()
        self._free_statuscb()
        if self.own and self.wrapped and gpgme.gpgme_release:
            gpgme.gpgme_release(self.wrapped)
            self.wrapped = None

    # Implement the context manager protocol.
    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        self.__del__()

    def op_keylist_all(self, *args, **kwargs):
        self.op_keylist_start(*args, **kwargs)
        key = self.op_keylist_next()
        while key:
            yield key
            key = self.op_keylist_next()
        self.op_keylist_end()

    def op_keylist_next(self):
        """Returns the next key in the list created
        by a call to op_keylist_start().  The object returned
        is of type Key."""
        ptr = gpgme.new_gpgme_key_t_p()
        try:
            errorcheck(gpgme.gpgme_op_keylist_next(self.wrapped, ptr))
            key = gpgme.gpgme_key_t_p_value(ptr)
        except errors.GPGMEError as excp:
            key = None
            if excp.getcode() != errors.EOF:
                raise excp
        gpgme.delete_gpgme_key_t_p(ptr)
        if key:
            key.__del__ = lambda self: gpgme.gpgme_key_unref(self)
            return key

    def get_key(self, fpr, secret=False):
        """Get a key given a fingerprint

        Keyword arguments:
        secret		-- to request a secret key

        Returns:
                        -- the matching key

        Raises:
        KeyError	-- if the key was not found
        GPGMEError	-- as signaled by the underlying library

        """
        ptr = gpgme.new_gpgme_key_t_p()

        try:
            errorcheck(gpgme.gpgme_get_key(self.wrapped, fpr, ptr, secret))
        except errors.GPGMEError as e:
            if e.getcode() == errors.EOF:
                raise errors.KeyNotFound(fpr)
            raise e

        key = gpgme.gpgme_key_t_p_value(ptr)
        gpgme.delete_gpgme_key_t_p(ptr)
        assert key
        key.__del__ = lambda self: gpgme.gpgme_key_unref(self)
        return key

    def op_trustlist_all(self, *args, **kwargs):
        self.op_trustlist_start(*args, **kwargs)
        trust = self.op_trustlist_next()
        while trust:
            yield trust
            trust = self.op_trustlist_next()
        self.op_trustlist_end()

    def op_trustlist_next(self):
        """Returns the next trust item in the list created
        by a call to op_trustlist_start().  The object returned
        is of type TrustItem."""
        ptr = gpgme.new_gpgme_trust_item_t_p()
        try:
            errorcheck(gpgme.gpgme_op_trustlist_next(self.wrapped, ptr))
            trust = gpgme.gpgme_trust_item_t_p_value(ptr)
        except errors.GPGMEError as excp:
            trust = None
            if excp.getcode() != errors.EOF:
                raise
        gpgme.delete_gpgme_trust_item_t_p(ptr)
        return trust

    def set_passphrase_cb(self, func, hook=None):
        """Sets the passphrase callback to the function specified by func.

        When the system needs a passphrase, it will call func with three args:
        hint, a string describing the key it needs the passphrase for;
        desc, a string describing the passphrase it needs;
        prev_bad, a boolean equal True if this is a call made after
        unsuccessful previous attempt.

        If hook has a value other than None it will be passed into the func
        as a forth argument.

        Please see the GPGME manual for more information.
        """
        if func is None:
            hookdata = None
        else:
            if hook is None:
                hookdata = (weakref.ref(self), func)
            else:
                hookdata = (weakref.ref(self), func, hook)
        gpgme.gpg_set_passphrase_cb(self, hookdata)

    def _free_passcb(self):
        if gpgme.gpg_set_passphrase_cb:
            self.set_passphrase_cb(None)

    def set_progress_cb(self, func, hook=None):
        """Sets the progress meter callback to the function specified by FUNC.
        If FUNC is None, the callback will be cleared.

        This function will be called to provide an interactive update
        of the system's progress.  The function will be called with
        three arguments, type, total, and current.  If HOOK is not
        None, it will be supplied as fourth argument.

        Please see the GPGME manual for more information.

        """
        if func is None:
            hookdata = None
        else:
            if hook is None:
                hookdata = (weakref.ref(self), func)
            else:
                hookdata = (weakref.ref(self), func, hook)
        gpgme.gpg_set_progress_cb(self, hookdata)

    def _free_progresscb(self):
        if gpgme.gpg_set_progress_cb:
            self.set_progress_cb(None)

    def set_status_cb(self, func, hook=None):
        """Sets the status callback to the function specified by FUNC.  If
        FUNC is None, the callback will be cleared.

        The function will be called with two arguments, keyword and
        args.  If HOOK is not None, it will be supplied as third
        argument.

        Please see the GPGME manual for more information.

        """
        if func is None:
            hookdata = None
        else:
            if hook is None:
                hookdata = (weakref.ref(self), func)
            else:
                hookdata = (weakref.ref(self), func, hook)
        gpgme.gpg_set_status_cb(self, hookdata)

    def _free_statuscb(self):
        if gpgme.gpg_set_status_cb:
            self.set_status_cb(None)

    @property
    def engine_info(self):
        """Configuration of the engine currently in use"""
        p = self.protocol
        infos = [i for i in self.get_engine_info() if i.protocol == p]
        assert len(infos) == 1
        return infos[0]

    def get_engine_info(self):
        """Get engine configuration

        Returns information about all configured and installed
        engines.

        Returns:
        infos		-- a list of engine infos

        """
        return gpgme.gpgme_ctx_get_engine_info(self.wrapped)

    def set_engine_info(self, proto, file_name=None, home_dir=None):
        """Change engine configuration

        Changes the configuration of the crypto engine implementing
        the protocol 'proto' for the context.

        Keyword arguments:
        file_name	-- engine program file name (unchanged if None)
        home_dir	-- configuration directory (unchanged if None)

        """
        self.ctx_set_engine_info(proto, file_name, home_dir)

    def wait(self, hang):
        """Wait for asynchronous call to finish. Wait forever if hang is True.
        Raises an exception on errors.

        Please read the GPGME manual for more information.

        """
        ptr = gpgme.new_gpgme_error_t_p()
        gpgme.gpgme_wait(self.wrapped, ptr, hang)
        status = gpgme.gpgme_error_t_p_value(ptr)
        gpgme.delete_gpgme_error_t_p(ptr)
        errorcheck(status)

    def op_edit(self, key, func, fnc_value, out):
        """Start key editing using supplied callback function

        Note: This interface is deprecated and will be removed with
        GPGME 1.8.  Please use .interact instead.  Furthermore, we
        implement this using gpgme_op_interact, so callbacks will get
        called with string keywords instead of numeric status
        messages.  Code that is using constants.STATUS_X or
        constants.status.X will continue to work, whereas code using
        magic numbers will break as a result.

        """
        warnings.warn(
            "Call to deprecated method op_edit.", category=DeprecationWarning)
        return self.interact(key, func, sink=out, fnc_value=fnc_value)


class Data(GpgmeWrapper):
    """Data buffer

    A lot of data has to be exchanged between the user and the crypto
    engine, like plaintext messages, ciphertext, signatures and
    information about the keys.  The technical details about
    exchanging the data information are completely abstracted by
    GPGME.  The user provides and receives the data via `gpgme_data_t'
    objects, regardless of the communication protocol between GPGME
    and the crypto engine in use.

    This Data class is the implementation of the GpgmeData objects.

    Please see the information about __init__ for instantiation.

    """

    _ctype = 'gpgme_data_t'
    _cprefix = 'gpgme_data_'

    def _errorcheck(self, name):
        """This function should list all functions returning gpgme_error_t"""
        # This list is compiled using
        #
        # $ grep -v '^gpgme_error_t ' obj/lang/python/python3.5-gpg/gpgme.h \
        #   | awk "/\(gpgme_data_t/ { printf (\"'%s',\\n\", \$2) } " \
        #   | sed "s/'\\*/'/"
        return name not in {
            'gpgme_data_read',
            'gpgme_data_write',
            'gpgme_data_seek',
            'gpgme_data_release',
            'gpgme_data_release_and_get_mem',
            'gpgme_data_get_encoding',
            'gpgme_data_get_file_name',
            'gpgme_data_set_flag',
            'gpgme_data_identify',
        }

    def __init__(self,
                 string=None,
                 file=None,
                 offset=None,
                 length=None,
                 cbs=None,
                 copy=True):
        """Initialize a new gpgme_data_t object.

        If no args are specified, make it an empty object.

        If string alone is specified, initialize it with the data
        contained there.

        If file, offset, and length are all specified, file must
        be either a filename or a file-like object, and the object
        will be initialized by reading the specified chunk from the file.

        If cbs is specified, it MUST be a tuple of the form:

        (read_cb, write_cb, seek_cb, release_cb[, hook])

        where the first four items are functions implementing reading,
        writing, seeking the data, and releasing any resources once
        the data object is deallocated.  The functions must match the
        following prototypes:

            def read(amount, hook=None):
                return <a b"bytes" object>

            def write(data, hook=None):
                return <the number of bytes written>

            def seek(offset, whence, hook=None):
                return <the new file position>

            def release(hook=None):
                <return value and exceptions are ignored>

        The functions may be bound methods.  In that case, you can
        simply use the 'self' reference instead of using a hook.

        If file is specified without any other arguments, then
        it must be a filename, and the object will be initialized from
        that file.

        """
        super(Data, self).__init__(None)
        self.data_cbs = None

        if cbs is not None:
            self.new_from_cbs(*cbs)
        elif string is not None:
            self.new_from_mem(string, copy)
        elif file is not None and offset is not None and length is not None:
            self.new_from_filepart(file, offset, length)
        elif file is not None:
            if util.is_a_string(file):
                self.new_from_file(file, copy)
            else:
                self.new_from_fd(file)
        else:
            self.new()

    def __del__(self):
        if not gpgme:
            # At interpreter shutdown, gpgme is set to NONE.
            return

        if self.wrapped is not None and gpgme.gpgme_data_release:
            gpgme.gpgme_data_release(self.wrapped)
            if self._callback_excinfo:
                gpgme.gpg_raise_callback_exception(self)
            self.wrapped = None
        self._free_datacbs()

    # Implement the context manager protocol.
    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        self.__del__()

    def _free_datacbs(self):
        self._data_cbs = None

    def new(self):
        tmp = gpgme.new_gpgme_data_t_p()
        errorcheck(gpgme.gpgme_data_new(tmp))
        self.wrapped = gpgme.gpgme_data_t_p_value(tmp)
        gpgme.delete_gpgme_data_t_p(tmp)

    def new_from_mem(self, string, copy=True):
        tmp = gpgme.new_gpgme_data_t_p()
        errorcheck(
            gpgme.gpgme_data_new_from_mem(tmp, string, len(string), copy))
        self.wrapped = gpgme.gpgme_data_t_p_value(tmp)
        gpgme.delete_gpgme_data_t_p(tmp)

    def new_from_file(self, filename, copy=True):
        tmp = gpgme.new_gpgme_data_t_p()
        try:
            errorcheck(gpgme.gpgme_data_new_from_file(tmp, filename, copy))
        except errors.GPGMEError as e:
            if e.getcode() == errors.INV_VALUE and not copy:
                raise ValueError("delayed reads are not yet supported")
            else:
                raise e
        self.wrapped = gpgme.gpgme_data_t_p_value(tmp)
        gpgme.delete_gpgme_data_t_p(tmp)

    def new_from_cbs(self, read_cb, write_cb, seek_cb, release_cb, hook=None):
        tmp = gpgme.new_gpgme_data_t_p()
        if hook is not None:
            hookdata = (weakref.ref(self), read_cb, write_cb, seek_cb,
                        release_cb, hook)
        else:
            hookdata = (weakref.ref(self), read_cb, write_cb, seek_cb,
                        release_cb)
        gpgme.gpg_data_new_from_cbs(self, hookdata, tmp)
        self.wrapped = gpgme.gpgme_data_t_p_value(tmp)
        gpgme.delete_gpgme_data_t_p(tmp)

    def new_from_filepart(self, file, offset, length):
        """This wraps the GPGME gpgme_data_new_from_filepart() function.
        The argument "file" may be:

        * a string specifying a file name, or
        * a file-like object supporting the fileno() and the mode attribute.

        """

        tmp = gpgme.new_gpgme_data_t_p()
        filename = None
        fp = None

        if util.is_a_string(file):
            filename = file
        else:
            fp = gpgme.fdopen(file.fileno(), file.mode)
            if fp is None:
                raise ValueError("Failed to open file from %s arg %s" % (str(
                    type(file)), str(file)))

        errorcheck(
            gpgme.gpgme_data_new_from_filepart(tmp, filename, fp, offset,
                                               length))
        self.wrapped = gpgme.gpgme_data_t_p_value(tmp)
        gpgme.delete_gpgme_data_t_p(tmp)

    def new_from_fd(self, file):
        """This wraps the GPGME gpgme_data_new_from_fd() function.  The
        argument "file" must be a file-like object, supporting the
        fileno() method.

        """
        tmp = gpgme.new_gpgme_data_t_p()
        errorcheck(gpgme.gpgme_data_new_from_fd(tmp, file.fileno()))
        self.wrapped = gpgme.gpgme_data_t_p_value(tmp)
        gpgme.delete_gpgme_data_t_p(tmp)

    def new_from_stream(self, file):
        """This wrap around gpgme_data_new_from_stream is an alias for
        new_from_fd() method since in python there's no difference
        between file stream and file descriptor."""
        self.new_from_fd(file)

    def new_from_estream(self, file):
        """This wrap around gpgme_data_new_from_estream is an alias for
        new_from_fd() method since in python there's no difference
        between file stream and file descriptor, but using fd broke."""
        self.new_from_stream(file)

    def write(self, buffer):
        """Write buffer given as string or bytes.

        If a string is given, it is implicitly encoded using UTF-8."""
        written = gpgme.gpgme_data_write(self.wrapped, buffer)
        if written < 0:
            if self._callback_excinfo:
                gpgme.gpg_raise_callback_exception(self)
            else:
                raise GPGMEError.fromSyserror()
        return written

    def read(self, size=-1):
        """Read at most size bytes, returned as bytes.

        If the size argument is negative or omitted, read until EOF is reached.

        Returns the data read, or the empty string if there was no data
        to read before EOF was reached."""

        if size == 0:
            return ''

        if size > 0:
            try:
                result = gpgme.gpgme_data_read(self.wrapped, size)
            except:
                if self._callback_excinfo:
                    gpgme.gpg_raise_callback_exception(self)
                else:
                    raise
            return result
        else:
            chunks = []
            while True:
                try:
                    result = gpgme.gpgme_data_read(self.wrapped, 4096)
                except:
                    if self._callback_excinfo:
                        gpgme.gpg_raise_callback_exception(self)
                    else:
                        raise
                if len(result) == 0:
                    break
                chunks.append(result)
            return b''.join(chunks)


def pubkey_algo_string(subkey):
    """Return short algorithm string

    Return a public key algorithm string (e.g. "rsa2048") for a given
    SUBKEY.

    Returns:
    algo      - a string

    """
    return gpgme.gpgme_pubkey_algo_string(subkey)


def pubkey_algo_name(algo):
    """Return name of public key algorithm

    Return the name of the public key algorithm for a given numeric
    algorithm id ALGO (cf. RFC4880).

    Returns:
    algo      - a string

    """
    return gpgme.gpgme_pubkey_algo_name(algo)


def hash_algo_name(algo):
    """Return name of hash algorithm

    Return the name of the hash algorithm for a given numeric
    algorithm id ALGO (cf. RFC4880).

    Returns:
    algo      - a string

    """
    return gpgme.gpgme_hash_algo_name(algo)


def get_protocol_name(proto):
    """Get protocol description

    Get the string describing protocol PROTO.

    Returns:
    proto     - a string

    """
    return gpgme.gpgme_get_protocol_name(proto)


def addrspec_from_uid(uid):
    """Return the address spec

    Return the addr-spec (cf. RFC2822 section 4.3) from a user id UID.

    Returns:
    addr_spec - a string

    """
    return gpgme.gpgme_addrspec_from_uid(uid)


def check_version(version=None):
    return gpgme.gpgme_check_version(version)


# check_version also makes sure that several subsystems are properly
# initialized, and it must be run at least once before invoking any
# other function.  We do it here so that the user does not have to do
# it unless she really wants to check for a certain version.
check_version()


def engine_check_version(proto):
    try:
        errorcheck(gpgme.gpgme_engine_check_version(proto))
        return True
    except errors.GPGMEError:
        return False


def get_engine_info():
    ptr = gpgme.new_gpgme_engine_info_t_p()
    try:
        errorcheck(gpgme.gpgme_get_engine_info(ptr))
        info = gpgme.gpgme_engine_info_t_p_value(ptr)
    except errors.GPGMEError:
        info = None
    gpgme.delete_gpgme_engine_info_t_p(ptr)
    return info


def set_engine_info(proto, file_name, home_dir=None):
    """Changes the default configuration of the crypto engine implementing
    the protocol 'proto'. 'file_name' is the file name of
    the executable program implementing this protocol. 'home_dir' is the
    directory name of the configuration directory (engine's default is
    used if omitted)."""
    errorcheck(gpgme.gpgme_set_engine_info(proto, file_name, home_dir))


def set_locale(category, value):
    """Sets the default locale used by contexts"""
    errorcheck(gpgme.gpgme_set_locale(None, category, value))


def wait(hang):
    """Wait for asynchronous call on any Context  to finish.
    Wait forever if hang is True.

    For finished anynch calls it returns a tuple (status, context):
        status  - status return by asnynchronous call.
        context - context which caused this call to return.

    Please read the GPGME manual of more information."""
    ptr = gpgme.new_gpgme_error_t_p()
    context = gpgme.gpgme_wait(None, ptr, hang)
    status = gpgme.gpgme_error_t_p_value(ptr)
    gpgme.delete_gpgme_error_t_p(ptr)
    if context is None:
        errorcheck(status)
    else:
        context = Context(context)
    return (status, context)



"""D-Bus exceptions."""

# Copyright (C) 2007 Collabora Ltd. <http://www.collabora.co.uk/>
#
# SPDX-License-Identifier: MIT
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use, copy,
# modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

__all__ = ('DBusException', 'MissingErrorHandlerException',
           'MissingReplyHandlerException', 'ValidationException',
           'IntrospectionParserException', 'UnknownMethodException',
           'NameExistsException')

from dbus._compat import is_py3


class DBusException(Exception):

    include_traceback = False
    """If True, tracebacks will be included in the exception message sent to
    D-Bus clients.

    Exceptions that are not DBusException subclasses always behave
    as though this is True. Set this to True on DBusException subclasses
    that represent a programming error, and leave it False on subclasses that
    represent an expected failure condition (e.g. a network server not
    responding)."""

    def __init__(self, *args, **kwargs):
        name = kwargs.pop('name', None)
        if name is not None or getattr(self, '_dbus_error_name', None) is None:
            self._dbus_error_name = name
        if kwargs:
            raise TypeError('DBusException does not take keyword arguments: %s'
                            % ', '.join(kwargs.keys()))
        Exception.__init__(self, *args)

    def __unicode__(self):
        """Return a unicode error"""
        # We can't just use Exception.__unicode__ because it chains up weirdly.
        # https://code.launchpad.net/~mvo/ubuntu/quantal/dbus-python/lp846044/+merge/129214
        if len(self.args) > 1:
            s = unicode(self.args)
        else:
            s = ''.join(self.args)

        if self._dbus_error_name is not None:
            return '%s: %s' % (self._dbus_error_name, s)
        else:
            return s

    def __str__(self):
        """Return a str error"""
        s = Exception.__str__(self)
        if self._dbus_error_name is not None:
            return '%s: %s' % (self._dbus_error_name, s)
        else:
            return s

    def get_dbus_message(self):
        if len(self.args) > 1:
            if is_py3:
                s = str(self.args)
            else:
                s = unicode(self.args)
        else:
            s = ''.join(self.args)

        if isinstance(s, bytes):
            return s.decode('utf-8', 'replace')

        return s

    def get_dbus_name(self):
        return self._dbus_error_name

class MissingErrorHandlerException(DBusException):

    include_traceback = True

    def __init__(self):
        DBusException.__init__(self, "error_handler not defined: if you define a reply_handler you must also define an error_handler")

class MissingReplyHandlerException(DBusException):

    include_traceback = True

    def __init__(self):
        DBusException.__init__(self, "reply_handler not defined: if you define an error_handler you must also define a reply_handler")

class ValidationException(DBusException):

    include_traceback = True

    def __init__(self, msg=''):
        DBusException.__init__(self, "Error validating string: %s"%msg)

class IntrospectionParserException(DBusException):

    include_traceback = True

    def __init__(self, msg=''):
        DBusException.__init__(self, "Error parsing introspect data: %s"%msg)

class UnknownMethodException(DBusException):

    include_traceback = True
    _dbus_error_name = 'org.freedesktop.DBus.Error.UnknownMethod'

    def __init__(self, method):
        DBusException.__init__(self, "Unknown method: %s"%method)

class NameExistsException(DBusException):

    include_traceback = True

    def __init__(self, name):
        DBusException.__init__(self, "Bus name already exists: %s"%name)



# This file was automatically generated by SWIG (http://www.swig.org).
# Version 4.0.2
#
# Do not make changes to this file unless you know what you are doing--modify
# the SWIG interface file instead.

from sys import version_info as _swig_python_version_info
if _swig_python_version_info < (2, 7, 0):
    raise RuntimeError("Python 2.7 or later required")

# Import the low-level C/C++ module
if __package__ or "." in __name__:
    from . import _selinux
else:
    import _selinux

try:
    import builtins as __builtin__
except ImportError:
    import __builtin__

def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except __builtin__.Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)


def _swig_setattr_nondynamic_instance_variable(set):
    def set_instance_attr(self, name, value):
        if name == "thisown":
            self.this.own(value)
        elif name == "this":
            set(self, name, value)
        elif hasattr(self, name) and isinstance(getattr(type(self), name), property):
            set(self, name, value)
        else:
            raise AttributeError("You cannot add instance attributes to %s" % self)
    return set_instance_attr


def _swig_setattr_nondynamic_class_variable(set):
    def set_class_attr(cls, name, value):
        if hasattr(cls, name) and not isinstance(getattr(cls, name), property):
            set(cls, name, value)
        else:
            raise AttributeError("You cannot add class attributes to %s" % cls)
    return set_class_attr


def _swig_add_metaclass(metaclass):
    """Class decorator for adding a metaclass to a SWIG wrapped class - a slimmed down version of six.add_metaclass"""
    def wrapper(cls):
        return metaclass(cls.__name__, cls.__bases__, cls.__dict__.copy())
    return wrapper


class _SwigNonDynamicMeta(type):
    """Meta class to enforce nondynamic attributes (no new attributes) for a class"""
    __setattr__ = _swig_setattr_nondynamic_class_variable(type.__setattr__)




import shutil
import os

DISABLED = -1
PERMISSIVE = 0
ENFORCING = 1

def restorecon(path, recursive=False, verbose=False, force=False, nthreads=1):
    """ Restore SELinux context on a given path

    Arguments:
    path -- The pathname for the file or directory to be relabeled.

    Keyword arguments:
    recursive -- Change files and directories file labels recursively (default False)
    verbose -- Show changes in file labels (default False)
    force -- Force reset of context to match file_context for customizable files,
    and the default file context, changing the user, role, range portion  as well
    as the type (default False)
    nthreads -- The number of threads to use during relabeling, or 0 to use as many
    threads as there are online CPU cores (default 1)
    """

    restorecon_flags = SELINUX_RESTORECON_IGNORE_DIGEST | SELINUX_RESTORECON_REALPATH
    if recursive:
        restorecon_flags |= SELINUX_RESTORECON_RECURSE
    if verbose:
        restorecon_flags |= SELINUX_RESTORECON_VERBOSE
    if force:
        restorecon_flags |= SELINUX_RESTORECON_SET_SPECFILE_CTX
    selinux_restorecon_parallel(os.path.expanduser(path), restorecon_flags, nthreads)

def chcon(path, context, recursive=False):
    """ Set the SELinux context on a given path """
    lsetfilecon(path, context)
    if recursive:
        for root, dirs, files in os.walk(path):
            for name in files + dirs:
                lsetfilecon(os.path.join(root, name), context)

def copytree(src, dest):
    """ An SELinux-friendly shutil.copytree method """
    shutil.copytree(src, dest)
    restorecon(dest, recursive=True)

def install(src, dest):
    """ An SELinux-friendly shutil.move method """
    shutil.move(src, dest)
    restorecon(dest, recursive=True)

class security_id(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    ctx = property(_selinux.security_id_ctx_get, _selinux.security_id_ctx_set)
    refcnt = property(_selinux.security_id_refcnt_get, _selinux.security_id_refcnt_set)

    def __init__(self):
        _selinux.security_id_swiginit(self, _selinux.new_security_id())
    __swig_destroy__ = _selinux.delete_security_id

# Register security_id in _selinux:
_selinux.security_id_swigregister(security_id)


def avc_sid_to_context(sid):
    return _selinux.avc_sid_to_context(sid)

def avc_sid_to_context_raw(sid):
    return _selinux.avc_sid_to_context_raw(sid)

def avc_context_to_sid(ctx):
    return _selinux.avc_context_to_sid(ctx)

def avc_context_to_sid_raw(ctx):
    return _selinux.avc_context_to_sid_raw(ctx)

def sidget(sid):
    return _selinux.sidget(sid)

def sidput(sid):
    return _selinux.sidput(sid)

def avc_get_initial_sid(name):
    return _selinux.avc_get_initial_sid(name)
class avc_entry_ref(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    ae = property(_selinux.avc_entry_ref_ae_get, _selinux.avc_entry_ref_ae_set)

    def __init__(self):
        _selinux.avc_entry_ref_swiginit(self, _selinux.new_avc_entry_ref())
    __swig_destroy__ = _selinux.delete_avc_entry_ref

# Register avc_entry_ref in _selinux:
_selinux.avc_entry_ref_swigregister(avc_entry_ref)

class avc_memory_callback(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    func_malloc = property(_selinux.avc_memory_callback_func_malloc_get, _selinux.avc_memory_callback_func_malloc_set)
    func_free = property(_selinux.avc_memory_callback_func_free_get, _selinux.avc_memory_callback_func_free_set)

    def __init__(self):
        _selinux.avc_memory_callback_swiginit(self, _selinux.new_avc_memory_callback())
    __swig_destroy__ = _selinux.delete_avc_memory_callback

# Register avc_memory_callback in _selinux:
_selinux.avc_memory_callback_swigregister(avc_memory_callback)

class avc_log_callback(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    func_log = property(_selinux.avc_log_callback_func_log_get, _selinux.avc_log_callback_func_log_set)
    func_audit = property(_selinux.avc_log_callback_func_audit_get, _selinux.avc_log_callback_func_audit_set)

    def __init__(self):
        _selinux.avc_log_callback_swiginit(self, _selinux.new_avc_log_callback())
    __swig_destroy__ = _selinux.delete_avc_log_callback

# Register avc_log_callback in _selinux:
_selinux.avc_log_callback_swigregister(avc_log_callback)

class avc_thread_callback(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    func_create_thread = property(_selinux.avc_thread_callback_func_create_thread_get, _selinux.avc_thread_callback_func_create_thread_set)
    func_stop_thread = property(_selinux.avc_thread_callback_func_stop_thread_get, _selinux.avc_thread_callback_func_stop_thread_set)

    def __init__(self):
        _selinux.avc_thread_callback_swiginit(self, _selinux.new_avc_thread_callback())
    __swig_destroy__ = _selinux.delete_avc_thread_callback

# Register avc_thread_callback in _selinux:
_selinux.avc_thread_callback_swigregister(avc_thread_callback)

class avc_lock_callback(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    func_alloc_lock = property(_selinux.avc_lock_callback_func_alloc_lock_get, _selinux.avc_lock_callback_func_alloc_lock_set)
    func_get_lock = property(_selinux.avc_lock_callback_func_get_lock_get, _selinux.avc_lock_callback_func_get_lock_set)
    func_release_lock = property(_selinux.avc_lock_callback_func_release_lock_get, _selinux.avc_lock_callback_func_release_lock_set)
    func_free_lock = property(_selinux.avc_lock_callback_func_free_lock_get, _selinux.avc_lock_callback_func_free_lock_set)

    def __init__(self):
        _selinux.avc_lock_callback_swiginit(self, _selinux.new_avc_lock_callback())
    __swig_destroy__ = _selinux.delete_avc_lock_callback

# Register avc_lock_callback in _selinux:
_selinux.avc_lock_callback_swigregister(avc_lock_callback)

AVC_OPT_UNUSED = _selinux.AVC_OPT_UNUSED
AVC_OPT_SETENFORCE = _selinux.AVC_OPT_SETENFORCE

def avc_init(msgprefix, mem_callbacks, log_callbacks, thread_callbacks, lock_callbacks):
    return _selinux.avc_init(msgprefix, mem_callbacks, log_callbacks, thread_callbacks, lock_callbacks)

def avc_open(opts, nopts):
    return _selinux.avc_open(opts, nopts)

def avc_cleanup():
    return _selinux.avc_cleanup()

def avc_reset():
    return _selinux.avc_reset()

def avc_destroy():
    return _selinux.avc_destroy()

def avc_has_perm_noaudit(ssid, tsid, tclass, requested, aeref, avd):
    return _selinux.avc_has_perm_noaudit(ssid, tsid, tclass, requested, aeref, avd)

def avc_has_perm(ssid, tsid, tclass, requested, aeref):
    return _selinux.avc_has_perm(ssid, tsid, tclass, requested, aeref)

def avc_audit(ssid, tsid, tclass, requested, avd, result):
    return _selinux.avc_audit(ssid, tsid, tclass, requested, avd, result)

def avc_compute_create(ssid, tsid, tclass):
    return _selinux.avc_compute_create(ssid, tsid, tclass)

def avc_compute_member(ssid, tsid, tclass):
    return _selinux.avc_compute_member(ssid, tsid, tclass)
AVC_CALLBACK_GRANT = _selinux.AVC_CALLBACK_GRANT
AVC_CALLBACK_TRY_REVOKE = _selinux.AVC_CALLBACK_TRY_REVOKE
AVC_CALLBACK_REVOKE = _selinux.AVC_CALLBACK_REVOKE
AVC_CALLBACK_RESET = _selinux.AVC_CALLBACK_RESET
AVC_CALLBACK_AUDITALLOW_ENABLE = _selinux.AVC_CALLBACK_AUDITALLOW_ENABLE
AVC_CALLBACK_AUDITALLOW_DISABLE = _selinux.AVC_CALLBACK_AUDITALLOW_DISABLE
AVC_CALLBACK_AUDITDENY_ENABLE = _selinux.AVC_CALLBACK_AUDITDENY_ENABLE
AVC_CALLBACK_AUDITDENY_DISABLE = _selinux.AVC_CALLBACK_AUDITDENY_DISABLE
AVC_CACHE_STATS = _selinux.AVC_CACHE_STATS
class avc_cache_stats(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    entry_lookups = property(_selinux.avc_cache_stats_entry_lookups_get, _selinux.avc_cache_stats_entry_lookups_set)
    entry_hits = property(_selinux.avc_cache_stats_entry_hits_get, _selinux.avc_cache_stats_entry_hits_set)
    entry_misses = property(_selinux.avc_cache_stats_entry_misses_get, _selinux.avc_cache_stats_entry_misses_set)
    entry_discards = property(_selinux.avc_cache_stats_entry_discards_get, _selinux.avc_cache_stats_entry_discards_set)
    cav_lookups = property(_selinux.avc_cache_stats_cav_lookups_get, _selinux.avc_cache_stats_cav_lookups_set)
    cav_hits = property(_selinux.avc_cache_stats_cav_hits_get, _selinux.avc_cache_stats_cav_hits_set)
    cav_probes = property(_selinux.avc_cache_stats_cav_probes_get, _selinux.avc_cache_stats_cav_probes_set)
    cav_misses = property(_selinux.avc_cache_stats_cav_misses_get, _selinux.avc_cache_stats_cav_misses_set)

    def __init__(self):
        _selinux.avc_cache_stats_swiginit(self, _selinux.new_avc_cache_stats())
    __swig_destroy__ = _selinux.delete_avc_cache_stats

# Register avc_cache_stats in _selinux:
_selinux.avc_cache_stats_swigregister(avc_cache_stats)


def avc_av_stats():
    return _selinux.avc_av_stats()

def avc_sid_stats():
    return _selinux.avc_sid_stats()

def avc_netlink_open(blocking):
    return _selinux.avc_netlink_open(blocking)

def avc_netlink_loop():
    return _selinux.avc_netlink_loop()

def avc_netlink_close():
    return _selinux.avc_netlink_close()

def selinux_status_open(fallback):
    return _selinux.selinux_status_open(fallback)

def selinux_status_close():
    return _selinux.selinux_status_close()

def selinux_status_updated():
    return _selinux.selinux_status_updated()

def selinux_status_getenforce():
    return _selinux.selinux_status_getenforce()

def selinux_status_policyload():
    return _selinux.selinux_status_policyload()

def selinux_status_deny_unknown():
    return _selinux.selinux_status_deny_unknown()
class context_s_t(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    ptr = property(_selinux.context_s_t_ptr_get, _selinux.context_s_t_ptr_set)

    def __init__(self):
        _selinux.context_s_t_swiginit(self, _selinux.new_context_s_t())
    __swig_destroy__ = _selinux.delete_context_s_t

# Register context_s_t in _selinux:
_selinux.context_s_t_swigregister(context_s_t)


def context_new(str):
    return _selinux.context_new(str)

def context_str(con):
    return _selinux.context_str(con)

def context_free(con):
    return _selinux.context_free(con)

def context_type_get(con):
    return _selinux.context_type_get(con)

def context_range_get(con):
    return _selinux.context_range_get(con)

def context_role_get(con):
    return _selinux.context_role_get(con)

def context_user_get(con):
    return _selinux.context_user_get(con)

def context_type_set(con, type):
    return _selinux.context_type_set(con, type)

def context_range_set(con, range):
    return _selinux.context_range_set(con, range)

def context_role_set(con, role):
    return _selinux.context_role_set(con, role)

def context_user_set(con, user):
    return _selinux.context_user_set(con, user)
SELINUX_DEFAULTUSER = _selinux.SELINUX_DEFAULTUSER

def get_ordered_context_list(user, fromcon):
    return _selinux.get_ordered_context_list(user, fromcon)

def get_ordered_context_list_with_level(user, level, fromcon):
    return _selinux.get_ordered_context_list_with_level(user, level, fromcon)

def get_default_context(user, fromcon):
    return _selinux.get_default_context(user, fromcon)

def get_default_context_with_level(user, level, fromcon):
    return _selinux.get_default_context_with_level(user, level, fromcon)

def get_default_context_with_role(user, role, fromcon):
    return _selinux.get_default_context_with_role(user, role, fromcon)

def get_default_context_with_rolelevel(user, role, level, fromcon):
    return _selinux.get_default_context_with_rolelevel(user, role, level, fromcon)

def query_user_context():
    return _selinux.query_user_context()

def manual_user_enter_context(user):
    return _selinux.manual_user_enter_context(user)

def selinux_default_type_path():
    return _selinux.selinux_default_type_path()

def get_default_type(role):
    return _selinux.get_default_type(role)
SELABEL_CTX_FILE = _selinux.SELABEL_CTX_FILE
SELABEL_CTX_MEDIA = _selinux.SELABEL_CTX_MEDIA
SELABEL_CTX_X = _selinux.SELABEL_CTX_X
SELABEL_CTX_DB = _selinux.SELABEL_CTX_DB
SELABEL_CTX_ANDROID_PROP = _selinux.SELABEL_CTX_ANDROID_PROP
SELABEL_CTX_ANDROID_SERVICE = _selinux.SELABEL_CTX_ANDROID_SERVICE
SELABEL_OPT_UNUSED = _selinux.SELABEL_OPT_UNUSED
SELABEL_OPT_VALIDATE = _selinux.SELABEL_OPT_VALIDATE
SELABEL_OPT_BASEONLY = _selinux.SELABEL_OPT_BASEONLY
SELABEL_OPT_PATH = _selinux.SELABEL_OPT_PATH
SELABEL_OPT_SUBSET = _selinux.SELABEL_OPT_SUBSET
SELABEL_OPT_DIGEST = _selinux.SELABEL_OPT_DIGEST
SELABEL_NOPT = _selinux.SELABEL_NOPT

def selabel_open(backend, opts, nopts):
    return _selinux.selabel_open(backend, opts, nopts)

def selabel_close(handle):
    return _selinux.selabel_close(handle)

def selabel_lookup(handle, key, type):
    return _selinux.selabel_lookup(handle, key, type)

def selabel_lookup_raw(handle, key, type):
    return _selinux.selabel_lookup_raw(handle, key, type)

def selabel_partial_match(handle, key):
    return _selinux.selabel_partial_match(handle, key)

def selabel_get_digests_all_partial_matches(rec, key, calculated_digest, xattr_digest):
    return _selinux.selabel_get_digests_all_partial_matches(rec, key, calculated_digest, xattr_digest)

def selabel_hash_all_partial_matches(rec, key, digest):
    return _selinux.selabel_hash_all_partial_matches(rec, key, digest)

def selabel_lookup_best_match(rec, key, type):
    return _selinux.selabel_lookup_best_match(rec, key, type)

def selabel_lookup_best_match_raw(rec, key, type):
    return _selinux.selabel_lookup_best_match_raw(rec, key, type)

def selabel_digest(rec, digest):
    return _selinux.selabel_digest(rec, digest)
SELABEL_SUBSET = _selinux.SELABEL_SUBSET
SELABEL_EQUAL = _selinux.SELABEL_EQUAL
SELABEL_SUPERSET = _selinux.SELABEL_SUPERSET
SELABEL_INCOMPARABLE = _selinux.SELABEL_INCOMPARABLE

def selabel_cmp(h1, h2):
    return _selinux.selabel_cmp(h1, h2)

def selabel_stats(handle):
    return _selinux.selabel_stats(handle)
SELABEL_X_PROP = _selinux.SELABEL_X_PROP
SELABEL_X_EXT = _selinux.SELABEL_X_EXT
SELABEL_X_CLIENT = _selinux.SELABEL_X_CLIENT
SELABEL_X_EVENT = _selinux.SELABEL_X_EVENT
SELABEL_X_SELN = _selinux.SELABEL_X_SELN
SELABEL_X_POLYPROP = _selinux.SELABEL_X_POLYPROP
SELABEL_X_POLYSELN = _selinux.SELABEL_X_POLYSELN
SELABEL_DB_DATABASE = _selinux.SELABEL_DB_DATABASE
SELABEL_DB_SCHEMA = _selinux.SELABEL_DB_SCHEMA
SELABEL_DB_TABLE = _selinux.SELABEL_DB_TABLE
SELABEL_DB_COLUMN = _selinux.SELABEL_DB_COLUMN
SELABEL_DB_SEQUENCE = _selinux.SELABEL_DB_SEQUENCE
SELABEL_DB_VIEW = _selinux.SELABEL_DB_VIEW
SELABEL_DB_PROCEDURE = _selinux.SELABEL_DB_PROCEDURE
SELABEL_DB_BLOB = _selinux.SELABEL_DB_BLOB
SELABEL_DB_TUPLE = _selinux.SELABEL_DB_TUPLE
SELABEL_DB_LANGUAGE = _selinux.SELABEL_DB_LANGUAGE
SELABEL_DB_EXCEPTION = _selinux.SELABEL_DB_EXCEPTION
SELABEL_DB_DATATYPE = _selinux.SELABEL_DB_DATATYPE

def selinux_restorecon(pathname, restorecon_flags):
    return _selinux.selinux_restorecon(pathname, restorecon_flags)

def selinux_restorecon_parallel(pathname, restorecon_flags, nthreads):
    return _selinux.selinux_restorecon_parallel(pathname, restorecon_flags, nthreads)
SELINUX_RESTORECON_IGNORE_DIGEST = _selinux.SELINUX_RESTORECON_IGNORE_DIGEST
SELINUX_RESTORECON_NOCHANGE = _selinux.SELINUX_RESTORECON_NOCHANGE
SELINUX_RESTORECON_SET_SPECFILE_CTX = _selinux.SELINUX_RESTORECON_SET_SPECFILE_CTX
SELINUX_RESTORECON_RECURSE = _selinux.SELINUX_RESTORECON_RECURSE
SELINUX_RESTORECON_VERBOSE = _selinux.SELINUX_RESTORECON_VERBOSE
SELINUX_RESTORECON_PROGRESS = _selinux.SELINUX_RESTORECON_PROGRESS
SELINUX_RESTORECON_REALPATH = _selinux.SELINUX_RESTORECON_REALPATH
SELINUX_RESTORECON_XDEV = _selinux.SELINUX_RESTORECON_XDEV
SELINUX_RESTORECON_ADD_ASSOC = _selinux.SELINUX_RESTORECON_ADD_ASSOC
SELINUX_RESTORECON_ABORT_ON_ERROR = _selinux.SELINUX_RESTORECON_ABORT_ON_ERROR
SELINUX_RESTORECON_SYSLOG_CHANGES = _selinux.SELINUX_RESTORECON_SYSLOG_CHANGES
SELINUX_RESTORECON_LOG_MATCHES = _selinux.SELINUX_RESTORECON_LOG_MATCHES
SELINUX_RESTORECON_IGNORE_NOENTRY = _selinux.SELINUX_RESTORECON_IGNORE_NOENTRY
SELINUX_RESTORECON_IGNORE_MOUNTS = _selinux.SELINUX_RESTORECON_IGNORE_MOUNTS
SELINUX_RESTORECON_MASS_RELABEL = _selinux.SELINUX_RESTORECON_MASS_RELABEL
SELINUX_RESTORECON_SKIP_DIGEST = _selinux.SELINUX_RESTORECON_SKIP_DIGEST
SELINUX_RESTORECON_CONFLICT_ERROR = _selinux.SELINUX_RESTORECON_CONFLICT_ERROR
SELINUX_RESTORECON_COUNT_ERRORS = _selinux.SELINUX_RESTORECON_COUNT_ERRORS

def selinux_restorecon_set_sehandle(hndl):
    return _selinux.selinux_restorecon_set_sehandle(hndl)

def selinux_restorecon_default_handle():
    return _selinux.selinux_restorecon_default_handle()

def selinux_restorecon_set_exclude_list():
    return _selinux.selinux_restorecon_set_exclude_list()

def selinux_restorecon_set_alt_rootpath(alt_rootpath):
    return _selinux.selinux_restorecon_set_alt_rootpath(alt_rootpath)
MATCH = _selinux.MATCH
NOMATCH = _selinux.NOMATCH
DELETED_MATCH = _selinux.DELETED_MATCH
DELETED_NOMATCH = _selinux.DELETED_NOMATCH
ERROR = _selinux.ERROR
class dir_xattr(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    directory = property(_selinux.dir_xattr_directory_get, _selinux.dir_xattr_directory_set)
    digest = property(_selinux.dir_xattr_digest_get, _selinux.dir_xattr_digest_set)
    result = property(_selinux.dir_xattr_result_get, _selinux.dir_xattr_result_set)
    next = property(_selinux.dir_xattr_next_get, _selinux.dir_xattr_next_set)

    def __init__(self):
        _selinux.dir_xattr_swiginit(self, _selinux.new_dir_xattr())
    __swig_destroy__ = _selinux.delete_dir_xattr

# Register dir_xattr in _selinux:
_selinux.dir_xattr_swigregister(dir_xattr)


def selinux_restorecon_xattr(pathname, xattr_flags, xattr_list):
    return _selinux.selinux_restorecon_xattr(pathname, xattr_flags, xattr_list)
SELINUX_RESTORECON_XATTR_RECURSE = _selinux.SELINUX_RESTORECON_XATTR_RECURSE
SELINUX_RESTORECON_XATTR_DELETE_NONMATCH_DIGESTS = _selinux.SELINUX_RESTORECON_XATTR_DELETE_NONMATCH_DIGESTS
SELINUX_RESTORECON_XATTR_DELETE_ALL_DIGESTS = _selinux.SELINUX_RESTORECON_XATTR_DELETE_ALL_DIGESTS
SELINUX_RESTORECON_XATTR_IGNORE_MOUNTS = _selinux.SELINUX_RESTORECON_XATTR_IGNORE_MOUNTS

def selinux_restorecon_get_skipped_errors():
    return _selinux.selinux_restorecon_get_skipped_errors()

def is_selinux_enabled():
    return _selinux.is_selinux_enabled()

def is_selinux_mls_enabled():
    return _selinux.is_selinux_mls_enabled()

def getcon():
    return _selinux.getcon()

def getcon_raw():
    return _selinux.getcon_raw()

def setcon(con):
    return _selinux.setcon(con)

def setcon_raw(con):
    return _selinux.setcon_raw(con)

def getpidcon(pid):
    return _selinux.getpidcon(pid)

def getpidcon_raw(pid):
    return _selinux.getpidcon_raw(pid)

def getprevcon():
    return _selinux.getprevcon()

def getprevcon_raw():
    return _selinux.getprevcon_raw()

def getpidprevcon(pid):
    return _selinux.getpidprevcon(pid)

def getpidprevcon_raw(pid):
    return _selinux.getpidprevcon_raw(pid)

def getexeccon():
    return _selinux.getexeccon()

def getexeccon_raw():
    return _selinux.getexeccon_raw()

def setexeccon(con):
    return _selinux.setexeccon(con)

def setexeccon_raw(con):
    return _selinux.setexeccon_raw(con)

def getfscreatecon():
    return _selinux.getfscreatecon()

def getfscreatecon_raw():
    return _selinux.getfscreatecon_raw()

def setfscreatecon(context):
    return _selinux.setfscreatecon(context)

def setfscreatecon_raw(context):
    return _selinux.setfscreatecon_raw(context)

def getkeycreatecon():
    return _selinux.getkeycreatecon()

def getkeycreatecon_raw():
    return _selinux.getkeycreatecon_raw()

def setkeycreatecon(context):
    return _selinux.setkeycreatecon(context)

def setkeycreatecon_raw(context):
    return _selinux.setkeycreatecon_raw(context)

def getsockcreatecon():
    return _selinux.getsockcreatecon()

def getsockcreatecon_raw():
    return _selinux.getsockcreatecon_raw()

def setsockcreatecon(context):
    return _selinux.setsockcreatecon(context)

def setsockcreatecon_raw(context):
    return _selinux.setsockcreatecon_raw(context)

def getfilecon(path):
    return _selinux.getfilecon(path)

def getfilecon_raw(path):
    return _selinux.getfilecon_raw(path)

def lgetfilecon(path):
    return _selinux.lgetfilecon(path)

def lgetfilecon_raw(path):
    return _selinux.lgetfilecon_raw(path)

def fgetfilecon(fd):
    return _selinux.fgetfilecon(fd)

def fgetfilecon_raw(fd):
    return _selinux.fgetfilecon_raw(fd)

def setfilecon(path, con):
    return _selinux.setfilecon(path, con)

def setfilecon_raw(path, con):
    return _selinux.setfilecon_raw(path, con)

def lsetfilecon(path, con):
    return _selinux.lsetfilecon(path, con)

def lsetfilecon_raw(path, con):
    return _selinux.lsetfilecon_raw(path, con)

def fsetfilecon(fd, con):
    return _selinux.fsetfilecon(fd, con)

def fsetfilecon_raw(fd, con):
    return _selinux.fsetfilecon_raw(fd, con)

def getpeercon(fd):
    return _selinux.getpeercon(fd)

def getpeercon_raw(fd):
    return _selinux.getpeercon_raw(fd)
class av_decision(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    allowed = property(_selinux.av_decision_allowed_get, _selinux.av_decision_allowed_set)
    decided = property(_selinux.av_decision_decided_get, _selinux.av_decision_decided_set)
    auditallow = property(_selinux.av_decision_auditallow_get, _selinux.av_decision_auditallow_set)
    auditdeny = property(_selinux.av_decision_auditdeny_get, _selinux.av_decision_auditdeny_set)
    seqno = property(_selinux.av_decision_seqno_get, _selinux.av_decision_seqno_set)
    flags = property(_selinux.av_decision_flags_get, _selinux.av_decision_flags_set)

    def __init__(self):
        _selinux.av_decision_swiginit(self, _selinux.new_av_decision())
    __swig_destroy__ = _selinux.delete_av_decision

# Register av_decision in _selinux:
_selinux.av_decision_swigregister(av_decision)

SELINUX_AVD_FLAGS_PERMISSIVE = _selinux.SELINUX_AVD_FLAGS_PERMISSIVE
class selinux_opt(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    type = property(_selinux.selinux_opt_type_get, _selinux.selinux_opt_type_set)
    value = property(_selinux.selinux_opt_value_get, _selinux.selinux_opt_value_set)

    def __init__(self):
        _selinux.selinux_opt_swiginit(self, _selinux.new_selinux_opt())
    __swig_destroy__ = _selinux.delete_selinux_opt

# Register selinux_opt in _selinux:
_selinux.selinux_opt_swigregister(selinux_opt)

class selinux_callback(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    func_log = property(_selinux.selinux_callback_func_log_get, _selinux.selinux_callback_func_log_set)
    func_audit = property(_selinux.selinux_callback_func_audit_get, _selinux.selinux_callback_func_audit_set)
    func_validate = property(_selinux.selinux_callback_func_validate_get, _selinux.selinux_callback_func_validate_set)
    func_setenforce = property(_selinux.selinux_callback_func_setenforce_get, _selinux.selinux_callback_func_setenforce_set)
    func_policyload = property(_selinux.selinux_callback_func_policyload_get, _selinux.selinux_callback_func_policyload_set)

    def __init__(self):
        _selinux.selinux_callback_swiginit(self, _selinux.new_selinux_callback())
    __swig_destroy__ = _selinux.delete_selinux_callback

# Register selinux_callback in _selinux:
_selinux.selinux_callback_swigregister(selinux_callback)

SELINUX_CB_LOG = _selinux.SELINUX_CB_LOG
SELINUX_CB_AUDIT = _selinux.SELINUX_CB_AUDIT
SELINUX_CB_VALIDATE = _selinux.SELINUX_CB_VALIDATE
SELINUX_CB_SETENFORCE = _selinux.SELINUX_CB_SETENFORCE
SELINUX_CB_POLICYLOAD = _selinux.SELINUX_CB_POLICYLOAD

def selinux_get_callback(type):
    return _selinux.selinux_get_callback(type)

def selinux_set_callback(type, cb):
    return _selinux.selinux_set_callback(type, cb)
SELINUX_ERROR = _selinux.SELINUX_ERROR
SELINUX_WARNING = _selinux.SELINUX_WARNING
SELINUX_INFO = _selinux.SELINUX_INFO
SELINUX_AVC = _selinux.SELINUX_AVC
SELINUX_POLICYLOAD = _selinux.SELINUX_POLICYLOAD
SELINUX_SETENFORCE = _selinux.SELINUX_SETENFORCE
SELINUX_TRANS_DIR = _selinux.SELINUX_TRANS_DIR

def security_compute_av(scon, tcon, tclass, requested, avd):
    return _selinux.security_compute_av(scon, tcon, tclass, requested, avd)

def security_compute_av_raw(scon, tcon, tclass, requested, avd):
    return _selinux.security_compute_av_raw(scon, tcon, tclass, requested, avd)

def security_compute_av_flags(scon, tcon, tclass, requested, avd):
    return _selinux.security_compute_av_flags(scon, tcon, tclass, requested, avd)

def security_compute_av_flags_raw(scon, tcon, tclass, requested, avd):
    return _selinux.security_compute_av_flags_raw(scon, tcon, tclass, requested, avd)

def security_compute_create(scon, tcon, tclass):
    return _selinux.security_compute_create(scon, tcon, tclass)

def security_compute_create_raw(scon, tcon, tclass):
    return _selinux.security_compute_create_raw(scon, tcon, tclass)

def security_compute_create_name(scon, tcon, tclass, objname):
    return _selinux.security_compute_create_name(scon, tcon, tclass, objname)

def security_compute_create_name_raw(scon, tcon, tclass, objname):
    return _selinux.security_compute_create_name_raw(scon, tcon, tclass, objname)

def security_compute_relabel(scon, tcon, tclass):
    return _selinux.security_compute_relabel(scon, tcon, tclass)

def security_compute_relabel_raw(scon, tcon, tclass):
    return _selinux.security_compute_relabel_raw(scon, tcon, tclass)

def security_compute_member(scon, tcon, tclass):
    return _selinux.security_compute_member(scon, tcon, tclass)

def security_compute_member_raw(scon, tcon, tclass):
    return _selinux.security_compute_member_raw(scon, tcon, tclass)

def security_compute_user(scon, username):
    return _selinux.security_compute_user(scon, username)

def security_compute_user_raw(scon, username):
    return _selinux.security_compute_user_raw(scon, username)

def security_validatetrans(scon, tcon, tclass, newcon):
    return _selinux.security_validatetrans(scon, tcon, tclass, newcon)

def security_validatetrans_raw(scon, tcon, tclass, newcon):
    return _selinux.security_validatetrans_raw(scon, tcon, tclass, newcon)

def security_load_policy(len):
    return _selinux.security_load_policy(len)

def security_get_initial_context(name):
    return _selinux.security_get_initial_context(name)

def security_get_initial_context_raw(name):
    return _selinux.security_get_initial_context_raw(name)

def selinux_mkload_policy(preservebools):
    return _selinux.selinux_mkload_policy(preservebools)

def selinux_init_load_policy():
    return _selinux.selinux_init_load_policy()
class SELboolean(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    name = property(_selinux.SELboolean_name_get, _selinux.SELboolean_name_set)
    value = property(_selinux.SELboolean_value_get, _selinux.SELboolean_value_set)

    def __init__(self):
        _selinux.SELboolean_swiginit(self, _selinux.new_SELboolean())
    __swig_destroy__ = _selinux.delete_SELboolean

# Register SELboolean in _selinux:
_selinux.SELboolean_swigregister(SELboolean)


def security_set_boolean_list(boolcnt, boollist, permanent):
    return _selinux.security_set_boolean_list(boolcnt, boollist, permanent)

def security_load_booleans(path):
    return _selinux.security_load_booleans(path)

def security_check_context(con):
    return _selinux.security_check_context(con)

def security_check_context_raw(con):
    return _selinux.security_check_context_raw(con)

def security_canonicalize_context(con):
    return _selinux.security_canonicalize_context(con)

def security_canonicalize_context_raw(con):
    return _selinux.security_canonicalize_context_raw(con)

def security_getenforce():
    return _selinux.security_getenforce()

def security_setenforce(value):
    return _selinux.security_setenforce(value)

def security_reject_unknown():
    return _selinux.security_reject_unknown()

def security_deny_unknown():
    return _selinux.security_deny_unknown()

def security_get_checkreqprot():
    return _selinux.security_get_checkreqprot()

def security_disable():
    return _selinux.security_disable()

def security_policyvers():
    return _selinux.security_policyvers()

def security_get_boolean_names():
    return _selinux.security_get_boolean_names()

def security_get_boolean_pending(name):
    return _selinux.security_get_boolean_pending(name)

def security_get_boolean_active(name):
    return _selinux.security_get_boolean_active(name)

def security_set_boolean(name, value):
    return _selinux.security_set_boolean(name, value)

def security_commit_booleans():
    return _selinux.security_commit_booleans()
class security_class_mapping(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    name = property(_selinux.security_class_mapping_name_get, _selinux.security_class_mapping_name_set)
    perms = property(_selinux.security_class_mapping_perms_get, _selinux.security_class_mapping_perms_set)

    def __init__(self):
        _selinux.security_class_mapping_swiginit(self, _selinux.new_security_class_mapping())
    __swig_destroy__ = _selinux.delete_security_class_mapping

# Register security_class_mapping in _selinux:
_selinux.security_class_mapping_swigregister(security_class_mapping)


def selinux_set_mapping(map):
    return _selinux.selinux_set_mapping(map)

def mode_to_security_class(mode):
    return _selinux.mode_to_security_class(mode)

def string_to_security_class(name):
    return _selinux.string_to_security_class(name)

def security_class_to_string(cls):
    return _selinux.security_class_to_string(cls)

def security_av_perm_to_string(tclass, perm):
    return _selinux.security_av_perm_to_string(tclass, perm)

def string_to_av_perm(tclass, name):
    return _selinux.string_to_av_perm(tclass, name)

def security_av_string(tclass, av):
    return _selinux.security_av_string(tclass, av)

def print_access_vector(tclass, av):
    return _selinux.print_access_vector(tclass, av)

def selinux_flush_class_cache():
    return _selinux.selinux_flush_class_cache()
MATCHPATHCON_BASEONLY = _selinux.MATCHPATHCON_BASEONLY
MATCHPATHCON_NOTRANS = _selinux.MATCHPATHCON_NOTRANS
MATCHPATHCON_VALIDATE = _selinux.MATCHPATHCON_VALIDATE

def set_matchpathcon_flags(flags):
    return _selinux.set_matchpathcon_flags(flags)

def matchpathcon_init(path):
    return _selinux.matchpathcon_init(path)

def matchpathcon_init_prefix(path, prefix):
    return _selinux.matchpathcon_init_prefix(path, prefix)

def matchpathcon_fini():
    return _selinux.matchpathcon_fini()

def realpath_not_final(name, resolved_path):
    return _selinux.realpath_not_final(name, resolved_path)

def matchpathcon(path, mode):
    return _selinux.matchpathcon(path, mode)

def matchpathcon_index(path, mode):
    return _selinux.matchpathcon_index(path, mode)

def matchpathcon_filespec_add(ino, specind, file):
    return _selinux.matchpathcon_filespec_add(ino, specind, file)

def matchpathcon_filespec_destroy():
    return _selinux.matchpathcon_filespec_destroy()

def matchpathcon_filespec_eval():
    return _selinux.matchpathcon_filespec_eval()

def matchpathcon_checkmatches(str):
    return _selinux.matchpathcon_checkmatches(str)

def matchmediacon(media):
    return _selinux.matchmediacon(media)

def selinux_getenforcemode():
    return _selinux.selinux_getenforcemode()

def selinux_boolean_sub(boolean_name):
    return _selinux.selinux_boolean_sub(boolean_name)

def selinux_getpolicytype():
    return _selinux.selinux_getpolicytype()

def selinux_policy_root():
    return _selinux.selinux_policy_root()

def selinux_set_policy_root(rootpath):
    return _selinux.selinux_set_policy_root(rootpath)

def selinux_current_policy_path():
    return _selinux.selinux_current_policy_path()

def selinux_binary_policy_path():
    return _selinux.selinux_binary_policy_path()

def selinux_failsafe_context_path():
    return _selinux.selinux_failsafe_context_path()

def selinux_removable_context_path():
    return _selinux.selinux_removable_context_path()

def selinux_default_context_path():
    return _selinux.selinux_default_context_path()

def selinux_user_contexts_path():
    return _selinux.selinux_user_contexts_path()

def selinux_file_context_path():
    return _selinux.selinux_file_context_path()

def selinux_file_context_homedir_path():
    return _selinux.selinux_file_context_homedir_path()

def selinux_file_context_local_path():
    return _selinux.selinux_file_context_local_path()

def selinux_file_context_subs_path():
    return _selinux.selinux_file_context_subs_path()

def selinux_file_context_subs_dist_path():
    return _selinux.selinux_file_context_subs_dist_path()

def selinux_homedir_context_path():
    return _selinux.selinux_homedir_context_path()

def selinux_media_context_path():
    return _selinux.selinux_media_context_path()

def selinux_virtual_domain_context_path():
    return _selinux.selinux_virtual_domain_context_path()

def selinux_virtual_image_context_path():
    return _selinux.selinux_virtual_image_context_path()

def selinux_lxc_contexts_path():
    return _selinux.selinux_lxc_contexts_path()

def selinux_x_context_path():
    return _selinux.selinux_x_context_path()

def selinux_sepgsql_context_path():
    return _selinux.selinux_sepgsql_context_path()

def selinux_openrc_contexts_path():
    return _selinux.selinux_openrc_contexts_path()

def selinux_openssh_contexts_path():
    return _selinux.selinux_openssh_contexts_path()

def selinux_snapperd_contexts_path():
    return _selinux.selinux_snapperd_contexts_path()

def selinux_systemd_contexts_path():
    return _selinux.selinux_systemd_contexts_path()

def selinux_contexts_path():
    return _selinux.selinux_contexts_path()

def selinux_securetty_types_path():
    return _selinux.selinux_securetty_types_path()

def selinux_booleans_subs_path():
    return _selinux.selinux_booleans_subs_path()

def selinux_booleans_path():
    return _selinux.selinux_booleans_path()

def selinux_customizable_types_path():
    return _selinux.selinux_customizable_types_path()

def selinux_users_path():
    return _selinux.selinux_users_path()

def selinux_usersconf_path():
    return _selinux.selinux_usersconf_path()

def selinux_translations_path():
    return _selinux.selinux_translations_path()

def selinux_colors_path():
    return _selinux.selinux_colors_path()

def selinux_netfilter_context_path():
    return _selinux.selinux_netfilter_context_path()

def selinux_path():
    return _selinux.selinux_path()

def selinux_check_access(scon, tcon, tclass, perm):
    return _selinux.selinux_check_access(scon, tcon, tclass, perm)

def selinux_check_passwd_access(requested):
    return _selinux.selinux_check_passwd_access(requested)

def checkPasswdAccess(requested):
    return _selinux.checkPasswdAccess(requested)

def selinux_check_securetty_context(tty_context):
    return _selinux.selinux_check_securetty_context(tty_context)

def set_selinuxmnt(mnt):
    return _selinux.set_selinuxmnt(mnt)

def selinuxfs_exists():
    return _selinux.selinuxfs_exists()

def fini_selinuxmnt():
    return _selinux.fini_selinuxmnt()

def setexecfilecon(filename, fallback_type):
    return _selinux.setexecfilecon(filename, fallback_type)

def is_context_customizable(scontext):
    return _selinux.is_context_customizable(scontext)

def selinux_trans_to_raw_context(trans):
    return _selinux.selinux_trans_to_raw_context(trans)

def selinux_raw_to_trans_context(raw):
    return _selinux.selinux_raw_to_trans_context(raw)

def selinux_raw_context_to_color(raw):
    return _selinux.selinux_raw_context_to_color(raw)

def getseuserbyname(linuxuser):
    return _selinux.getseuserbyname(linuxuser)

def getseuser(username, service):
    return _selinux.getseuser(username, service)

def selinux_file_context_cmp(a, b):
    return _selinux.selinux_file_context_cmp(a, b)

def selinux_file_context_verify(path, mode):
    return _selinux.selinux_file_context_verify(path, mode)

def selinux_lsetfilecon_default(path):
    return _selinux.selinux_lsetfilecon_default(path)

def selinux_reset_config():
    return _selinux.selinux_reset_config()





# Copyright 2016, Tresys Technology, LLC
# Copyright 2018, Chris PeBenito <pebenito@ieee.org>
#
# SPDX-License-Identifier: LGPL-2.1-only
#
from collections import defaultdict
from typing import NamedTuple

from ..policyrep import SELinuxPolicy, Boolean

from .descriptors import DiffResultDescriptor
from .difference import Difference, SymbolWrapper
from .typing import SymbolCache


_bool_cache: SymbolCache[Boolean] = defaultdict(dict)


class ModifiedBoolean(NamedTuple):

    """Difference details for a modified Boolean."""

    added_state: bool
    removed_state: bool


def boolean_wrapper(policy: SELinuxPolicy, boolean: Boolean) -> SymbolWrapper[Boolean]:
    """
    Wrap booleans from the specified policy.

    This caches results to prevent duplicate wrapper
    objects in memory.
    """
    try:
        return _bool_cache[policy][boolean]
    except KeyError:
        b = SymbolWrapper(boolean)
        _bool_cache[policy][boolean] = b
        return b


class BooleansDifference(Difference):

    """Determine the difference in type attributes between two policies."""

    added_booleans = DiffResultDescriptor("diff_booleans")
    removed_booleans = DiffResultDescriptor("diff_booleans")
    modified_booleans = DiffResultDescriptor("diff_booleans")

    def diff_booleans(self) -> None:
        """Generate the difference in type attributes between the policies."""

        self.log.info("Generating Boolean differences from {0.left_policy} to {0.right_policy}".
                      format(self))

        self.added_booleans, self.removed_booleans, matched_booleans = \
            self._set_diff(
                (SymbolWrapper(b) for b in self.left_policy.bools()),
                (SymbolWrapper(b) for b in self.right_policy.bools()))

        self.modified_booleans = dict()

        for left_boolean, right_boolean in matched_booleans:
            # Criteria for modified booleans
            # 1. change to default state
            if left_boolean.state != right_boolean.state:
                self.modified_booleans[left_boolean] = ModifiedBoolean(right_boolean.state,
                                                                       left_boolean.state)

    #
    # Internal functions
    #
    def _reset_diff(self) -> None:
        """Reset diff results on policy changes."""
        self.log.debug("Resetting Boolean differences")
        self.added_booleans = None
        self.removed_booleans = None
        self.modified_booleans = None
